"""
.. module:: sshagent
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`SshAgent` class and associated diagnostic.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Callable, Optional, Sequence, Union, Tuple
from types import TracebackType

import os
import re
import socket
import stat
import threading
import time
import weakref

from datetime import datetime, timedelta

from akit.aspects import ActionPattern, AspectsCmd, LoggingPattern, DEFAULT_CMD_ASPECTS
from akit.exceptions import AKitCommandError, AKitInvalidConfigError, AKitNotOverloadedError, AKitSemanticError, AKitTimeoutError
from akit.interfaces.icommandrunner import ICommandRunner
from akit.timeouts import TimeoutContext

from akit.xformatting import indent_lines, format_command_result
from akit.xlogging.foundations import getAutomatonKitLogger
from akit.xlogging.scopemonitoring import MonitoredScope

from akit.interop.credentials.sshcredential import SshCredential
from akit.interop.landscaping.landscapedeviceextension import LandscapeDeviceExtension

import paramiko

logger = getAutomatonKitLogger()

DEFAULT_SSH_TIMEOUT = 300
DEFAULT_SSH_RETRY_INTERVAL = .5

DEFAULT_FAILURE_LABEL = "Failure"
DEFAULT_SUCCESS_LABEL = "Success"

INTERACTIVE_PROMPT="PROMPT%:%:%"
INTERACTIVE_PROMPT_BYTES = INTERACTIVE_PROMPT.encode('utf-8')

TEMPLATE_COMMAND_FAILURE = "RUNCMD%s: {} running CMD=%s{}STDOUT:{}%s{}STDERR:{}%s".format(DEFAULT_FAILURE_LABEL, *([os.linesep] * 4))
TEMPLATE_COMMAND_SUCCESS = "RUNCMD%s: {} running CMD=%s{}STDOUT:{}%s{}STDERR:{}%s".format(DEFAULT_SUCCESS_LABEL, *([os.linesep] * 4))

#                    PERMS          LINKS   OWNER       GROUPS    SIZE   MONTH  DAY  TIME    NAME
#                  rwxrwxrwx         24      myron      myron     4096   Jul    4   00:37  PCBs
REGEX_DIRECTORY_ENTRY = re.compile(r"([\S]+)[\s]+([0-9]+)[\s]+([\S]+)[\s]+([\S]+)[\s]+([0-9]+)[\s]+([A-za-z]+[\s]+[0-9]+[\s]+[0-9:]+)[\s]+([\S\s]+)")
def lookup_entry_type(pdir: str, ename: str, finfo: paramiko.SFTPAttributes) -> str:
    """
        Determines the entry type labelto assign to a file entry based on the st_mode of the file information.

        :param pdir: The parent directory of the file, passed for logging purposes.
        :param ename: The entry name of the file in the directory listing.
        :param finfo: The file information as returned by lstat on the file entry.

        :returns: A label indicating the type of directory entry.
    """
    etype = None
    if stat.S_ISREG(finfo.st_mode):
        etype = "file"
    elif stat.S_ISDIR(finfo.st_mode):
        etype = "dir"
    elif stat.S_ISLNK(finfo.st_mode):
        etype = "link"
    elif stat.S_ISCHR(finfo.st_mode):
        etype = "char"
    elif stat.S_ISBLK(finfo.st_mode):
        etype = "block"
    else:
        print ("ERROR: Unknown entry type. pdir=%s, ename=%s" % (pdir, ename))

    return etype

def primitive_list_directory(ssh_client: paramiko.SSHClient, directory: str) -> dict:
    """
        Uses a primitive method to create a list of the files and folders in a directory
        by running the 'ls -al' commands on the directory.  This is required if the SSH
        server does not support sftp.

        :param ssh_client: A :class:`paramiko.SSHClient` object that has already been connected
                           to the remote server.
        :param directory: The remote directory to get a list catalog for.

        :returns: A dictionary with the information about the items in the target directory.
    """
    entries = None

    ls_cmd = "ls -al %s" % directory

    status, stdout, stderr = ssh_execute_command(ssh_client, ls_cmd)

    if status == 0:
        entries = primitive_parse_directory_listing(directory, stdout)
    else:
        # Allow permission denied errors
        if stderr.find("Permission denied") < 0:
            errmsg = "Error attempting to get a primitive directory listing.\n    CMD=%s\n    STDERR:\n%s" % (ls_cmd, stderr)
            raise Exception(errmsg)

    return entries

def primitive_list_tree(ssh_client: paramiko.SSHClient, treeroot: str, max_depth: int=1) -> dict:
    """
        Uses a primitive method to create an information tree about the files and folders
        in a directory tree by running 'ls -al' commands on the directory tree.  This is
        required if the SSH server does not support sftp.

        :param ssh_client: A :class:`paramiko.SSHClient` object that has already been connected
                           to the remote server.
        :param treeroot: The remote root directory to get a directory tree catalog for.
        :param max_depth: The maximum descent depth to go to when building the tree catalog.

        :returns: A dictionary with the information about the root directory and its children.
    """

    level_items = primitive_list_directory(ssh_client, treeroot)

    if max_depth > 1:
        if not treeroot.endswith("/"):
            treeroot = treeroot + "/"

        for nxtitem in level_items.values():
            nxtname = nxtitem["name"]
            if nxtitem["type"] == "dir":
                nxtroot = treeroot + nxtname
                nxtchildren = primitive_list_tree_recurse(ssh_client, nxtroot, max_depth - 1)
                nxtitem.update(nxtchildren)

    tree_info = {
        "name": treeroot,
        "items": level_items
    }

    return tree_info

def primitive_list_tree_recurse(ssh_client: paramiko.SSHClient, targetdir: str, remaining: int) -> dict:
    """
        Method which is called recursively to list the items in a directory and to form a tree
        of information about the files under a room directory.  This method utilizes primitive
        shell commands in order to get directory information and is used when the SSH server on a
        device does not support FTP based access or file queries

        :params ssh_client: A paramiko.SSHClient to use for running commands to discover the information
                            about a directory.
        :params targetdir: The target directory of the local tree to scan.
        :params remaining: The remaining levels to descend

        :returns: A dictionary with the information about the target directory and its children.
    """

    level_items = primitive_list_directory(ssh_client, targetdir)

    if remaining > 1:
        if not targetdir.endswith("/"):
            targetdir = targetdir + "/"

        for nxtitem in level_items.values():
            nxtname = nxtitem["name"]
            if nxtitem["type"] == "dir":
                nxttarget = targetdir + nxtname
                nxtchildren = primitive_list_tree_recurse(ssh_client, nxttarget, remaining - 1)
                nxtitem.update(nxtchildren)

    children_info = {
        "items": level_items
    }

    return children_info

def primitive_parse_directory_listing(listing_dir:str, content: str) -> dict:
    """
        Method which is called parse the information associated with a directory listing.

        :params listing_dir: The directory path associated with the directory listing content being parsed.
        :params content: The directory listing content to be parsed.

        :returns: A dictionary with the directory entry items from a directoy based on the directory listing content being parsed.
    """
    entries = {}

    listing_lines = content.splitlines(False)
    for nxtline in listing_lines:
        nxtline = nxtline.strip()

        mobj = REGEX_DIRECTORY_ENTRY.match(nxtline)
        if mobj is not None:
            mgroups = mobj.groups()

            name = mgroups[6]
            if name == "." or name == "..":
                continue

            perms = mgroups[0]

            perms_tc = perms[0]
            etype = None
            if perms_tc == '-':
                etype = "file"
            elif perms_tc == 'd':
                etype = "dir"
            elif perms_tc == 'l':
                etype = "link"
                name, _ = name.split(" -> ")
            elif perms_tc == 'c':
                etype = "char"
            elif perms_tc == 'b':
                etype = "block"
            else:
                raise Exception("Unknown directory listing entry type. dir=%s tc=%s" % (listing_dir, perms_tc))

            dentry = {
                "name": name,
                "perms": perms,
                "type": etype,
                "owner": mgroups[2],
                "group": mgroups[3],
                "size": mgroups[4],
                "modified": mgroups[5]
            }

            entries[name] = dentry

    return entries

def primitive_directory_exists(ssh_client: paramiko.SSHClient, remotedir: str) -> bool:
    """
        Checks to see if the specified file exists.

        :param ssh_client: A paramiko.SSHClient to run commands over.
        :param remotedir: The path of the remote directory to check for existance.

        :returns: A boolean indicating if the file exists.
    """
    exists = False

    existscmd = 'test -d {0} && echo exists'.format(remotedir)
    status, stdout, stderr = ssh_execute_command(ssh_client, existscmd)

    if status == 0:
        exists = stdout.strip() == 'exists'

    return exists

def primitive_file_exists(ssh_client: paramiko.SSHClient, remotepath: str) -> bool:
    """
        Checks to see if the specified file exists.

        :param ssh_client: A paramiko.SSHClient to run commands over.
        :param remotepath: The path of the remote file to check for existance.

        :returns: A boolean indicating if the file exists.
    """
    exists = False

    existscmd = 'test -f {0} && echo exists'.format(remotepath)
    status, stdout, stderr = ssh_execute_command(ssh_client, existscmd)

    if status == 0:
        exists = stdout.strip() == 'exists'

    return exists

def primitive_file_pull(ssh_client: paramiko.SSHClient, remotepath: str, localpath: str):
    """
        Pulls the contents of a file to a local path using a primitive 'cat' command.

        :param ssh_client: A paramiko.SSHClient to run commands over.
        :param remotepath: The path to the file on the remote machine.
        :param localpath: The local path to a file to write the contents of the remote file too.
    """
    copycmd = "cat %s" % remotepath

    status, stdout, stderr = ssh_execute_command(ssh_client, copycmd, decode=False)

    if status != 0:
        stderr = stderr.decode()
        errmsg_lines = [
            "Error pulling file using command. cmd=%s" % copycmd,
            "STDERR:",
            indent_lines(stderr, 1)
        ]
        raise Exception(os.linesep.join(errmsg_lines))

    with open(localpath, 'wb') as lfile:
        lfile.write(stdout)

    return

def primitive_file_push(ssh_client: paramiko.SSHClient, localpath: str, remotepath: str, read_size: int=1024):
    """
        Pushes the contents of a local file to a file on a remote machine at the path specified by remotepath.

        :param ssh_client: A paramiko.SSHClient to run commands over.
        :param localpath: The local path to a file to read contents from.
        :param remotepath: The path to file on the remote machine to write the contents of the local file too.
    """
    cat_cmd = "cat > {}".format(remotepath)

    with open(localpath, "rb") as file:
        content = file.read()
        status, stdout, stderr = ssh_execute_command(ssh_client, command=cat_cmd, chunk_size=read_size, input=content, decode=False)
        if status != 0:
            errmsg = "Error pushing file={} to remote path.".format(remotepath)
            fmt_errmsg = format_command_result(errmsg, status, stdout.decode(), stderr.decode())
            raise AKitCommandError(fmt_errmsg)

    return

def sftp_list_directory(sftp, directory, userlookup, grouplookup) -> dict:
    """
        Gets a listing of the directory entries of a remote directory.

        :param sftp: A paramiko.SFTPClient that is connected to a remote machine.
        :param directory: The remote directory to get the directory listing for.
        :param userlookup: A dictionary to utilize in order to lookup the user associated with a directory item.
        :param grouplookup: A dictionary to utilize in order to lookup the group associated with a directory item.

        :returns: A dictionary with the information about the items in the target directory.
    """
    entries = {}

    try:
        sftp.chdir(directory)
        directory_items = sftp.listdir()

        for nname in directory_items:

            try:
                ninfo = sftp.lstat(nname)
                perms = ""
                dentry = {
                    "name": nname,
                    "perms": perms,
                    "type": lookup_entry_type(directory, nname, ninfo),
                    "owner": userlookup(ninfo.st_uid),
                    "group": grouplookup(ninfo.st_gid),
                    "size": ninfo.st_size,
                    "accessed": ninfo.st_atime,
                    "modified": ninfo.st_mtime
                }

                entries[nname] = dentry
            except PermissionError:
                logger.exception("Unable to perform lstat on remote directory item %r." % nname)
    except PermissionError:
        logger.exception("Unable to perform listdir on remote directory %r." % directory)

    return entries

def sftp_list_tree(sftp: paramiko.SFTPClient, treeroot: str, userlookup: Callable[[int, Optional[bool]], str], grouplookup: Callable[[int, Optional[bool]], str], max_depth: int=1) -> dict:
    """
        Gets a tree of the directory entries of a remote directory.

        :param sftp: A paramiko.SFTPClient that is connected to a remote machine.
        :param treeroot: The root directory to get the directory listing for.
        :param userlookup: A dictionary to utilize in order to lookup the user associated with a directory item.
        :param grouplookup: A dictionary to utilize in order to lookup the group associated with a directory item.
        :param max_depth: The depth to which to create a directory tree.

        :returns: A dictionary with the information about the root directory and its children.
    """
    level_items = sftp_list_directory(sftp, treeroot, userlookup, grouplookup)

    if max_depth > 1:
        if not treeroot.endswith("/"):
            treeroot = treeroot + "/"

        for nxtitem in level_items.values():
            nxtname = nxtitem["name"]
            if nxtitem["type"] == "dir":
                nxtroot = treeroot + nxtname
                nxtchildren = sftp_list_tree_recurse(sftp, nxtroot, userlookup, grouplookup, max_depth - 1)
                nxtitem.update(nxtchildren)

    tree_info = {
        "name": treeroot,
        "items": level_items
    }

    return tree_info

def sftp_list_tree_recurse(sftp: paramiko.SFTPClient, targetdir: str, userlookup: Callable[[int, Optional[bool]], str], grouplookup: Callable[[int, Optional[bool]], str], remaining: int) -> dict:
    """
        Gets a listing of the directory passed and then recursively calls itself to create sub directory listings up to the remaining depth.

        :param sftp: A paramiko.SFTPClient that is connected to a remote machine.
        :param targetdir: The target directory to get the directory tree listing for.
        :param userlookup: A dictionary to utilize in order to lookup the user associated with a directory item.
        :param grouplookup: A dictionary to utilize in order to lookup the group associated with a directory item.
        :param max_depth: The depth to which to create a directory tree.

        :returns: A dictionary with the information about the target directory and its children.
    """
    level_items = sftp_list_directory(sftp, targetdir, userlookup, grouplookup)

    if remaining > 1:
        if not targetdir.endswith("/"):
            targetdir = targetdir + "/"

        for nxtitem in level_items.values():
            nxtname = nxtitem["name"]
            if nxtitem["type"] == "dir":
                nxtroot = targetdir + nxtname
                nxtchildren = sftp_list_tree_recurse(sftp, nxtroot, userlookup, grouplookup, remaining - 1)
                nxtitem.update(nxtchildren)

    children_info = {
        "items": level_items
    }

    return children_info

def ssh_execute_command(ssh_client: paramiko.SSHClient, command: str, pty_params=None,
                        inactivity_timeout: float=DEFAULT_SSH_TIMEOUT, inactivity_interval: float=DEFAULT_SSH_RETRY_INTERVAL,
                        chunk_size: int=1024, input: Optional[str]=None, decode=True) -> Tuple[int, str, str]:
    """
        Runs a command on a remote server using the specified ssh_client.  We implement our own version of ssh_execute_command
        in order to have better control over the timeouts and to make sure all the checks are sequenced properly in order
        to prevent SSH lockups.

        :param ssh_client: The :class:`paramiko.SSHClient` object to utilize when running the command.
        :param command: The commandline to run.
        :param inactivity_timeout: The timeout for the channel for the ssh transaction.
        :param inactivity_interval: The retry interval to wait for the channel to have data ready.
        :param chunk_size: The size of the chunks that are read during receive operations

        :returns: A tuple with the command result code, the standard output and the standard error output.
    """
    status = None
    stdout_buffer = bytearray()
    stderr_buffer = bytearray()

    start_time = time.time()
    end_time = start_time + inactivity_timeout

    channel = ssh_client.get_transport().open_session(timeout=inactivity_timeout)

    if pty_params is not None:
        channel.get_pty(**pty_params)

    channel.exec_command(command)

    input_length = 0
    if input is not None:
        if isinstance(input, str):
            input = input.encode()
        input_length = len(input)
    input_pos = 0
    input_shutdown = False

    while True:
        # Grab exit status ready before checking to see if stdout_ready or stderr_ready are True
        exit_status_ready = channel.exit_status_ready()
        stdout_ready = channel.recv_ready()
        stderr_ready = channel.recv_stderr_ready()

        # If stdout_ready or stderr_ready are true, we read this round
        if stdout_ready or stderr_ready:
            if stdout_ready:
                rcv_data = channel.recv(chunk_size)
                stdout_buffer.extend(rcv_data)
            if stderr_ready:
                rcv_data = channel.recv_stderr(chunk_size)
                stderr_buffer.extend(rcv_data)

            # We only want to timeout if there is inactivity
            start_time = time.time()
            end_time = start_time + inactivity_timeout

        # If there is data left to send and the send channel is ready
        # for input, send a chunk from the input buffer 
        elif input_length > 0 and input_pos < input_length and channel.send_ready():

            send_size = chunk_size
            avail_length = input_length - input_pos
            if avail_length < chunk_size:
                send_size = avail_length

            send_chunk = input[input_pos: input_pos + send_size]
            sent_size = channel.send(send_chunk)

            input_pos = input_pos + sent_size

        # If we have written everything there is to write, and writting to the
        # channel has not been shutdown, then shutdown writing.
        elif input is not None and not input_shutdown and input_pos >= input_length:
            channel.shutdown_write()
            input_shutdown = True

        # If we did not read any data this round and our exit status was ready, its time to exit
        elif exit_status_ready:
            status = channel.recv_exit_status()
            break
        # We didn't have any data to read and our exit status was not ready, go to sleep for interval
        else:
            now_time = time.time()
            if now_time > end_time:
                diff_time = now_time - start_time
                peername = ssh_client.get_transport().sock.getpeername()
                err_msg = "Timeout while executing SSH command.\n  peer=%r  command=%s\n  start=%r end=%r now=%r diff=%r" % (
                    peername, command, start_time, end_time, now_time, diff_time)
                raise TimeoutError(err_msg)

            # If we did not timeout, go to sleep for the interval before we make another attempt to read
            time.sleep(inactivity_interval)

        # End while True

    stdout = stdout_buffer
    if decode:
        stdout = stdout_buffer.decode()
    del stdout_buffer

    stderr = stderr_buffer
    if decode:
        stderr = stderr_buffer.decode()
    del stderr_buffer

    return status, stdout, stderr

def ssh_execute_command_in_channel(ssh_channel: paramiko.Channel, command: str, read_timeout: float=2, inactivity_timeout: float=DEFAULT_SSH_TIMEOUT, inactivity_interval: float=DEFAULT_SSH_RETRY_INTERVAL, chunk_size: int=1024, status_check=False) -> Tuple[int, str, str]:
    """
        Runs a command on a remote server using the specified ssh_client.  We implement our own version of ssh_execute_command
        in order to have better control over the timeouts and to make sure all the checks are sequenced properly in order
        to prevent SSH lockups.

        :param ssh_client: The :class:`paramiko.SSHClient` object to utilize when running the command.
        :param command: The commandline to run.
        :param inactivity_timeout: The timeout for the channel for the ssh transaction.
        :param inactivity_interval: The retry interval to wait for the channel to have data ready.
        :param chunk_size: The size of the chunks that are read during receive operations

        :returns: A tuple with the command result code, the standard output and the standard error output.
    """
    status = None
    stdout_buffer = bytearray()
    stderr_buffer = bytearray()

    start_time = time.time()
    end_time = start_time + inactivity_timeout

    ssh_channel.sendall(command + "\n")

    stdout_pipe_timeout = 0
    stderr_pipe_timeout = 0
    while True:
        if stdout_pipe_timeout < 2:
            try:
                rcv_data = ssh_channel.in_buffer.read(chunk_size, read_timeout)
                stdout_buffer.extend(rcv_data)
                if rcv_data.endswith(INTERACTIVE_PROMPT_BYTES):
                    stdout_pipe_timeout = 2
            except paramiko.buffered_pipe.PipeTimeout:
                stdout_pipe_timeout += 1

        if stderr_pipe_timeout < 2:
            try:
                rcv_data = ssh_channel.in_stderr_buffer.read(chunk_size, read_timeout)
                stderr_buffer.extend(rcv_data)
            except paramiko.buffered_pipe.PipeTimeout:
                stderr_pipe_timeout += 1

        if stdout_pipe_timeout > 1 and stderr_pipe_timeout > 1:
            break

        now_time = time.time()
        if now_time > end_time:
            errmsg = "Timeout while running command={}".format(command)
            raise AKitTimeoutError(errmsg)
        # End while True

    if not status_check:
        _, status_stdout, _ = ssh_execute_command_in_channel(ssh_channel, "echo $?", read_timeout=read_timeout, inactivity_timeout=inactivity_timeout, inactivity_interval=inactivity_interval, chunk_size=chunk_size, status_check=True)
        status = int(status_stdout)

    stdout = stdout_buffer.decode()
    del stdout_buffer

    if not stdout.startswith(command):
        errmsg = "The first line should have been an echo of our command."
        raise Exception(errmsg)

    if not stdout.endswith(INTERACTIVE_PROMPT):
        errmsg = "The first line should have been an echo of our command."
        raise Exception(errmsg)

    stdout = stdout[len(command):-len(INTERACTIVE_PROMPT)].strip()

    stderr = stderr_buffer.decode()
    del stderr_buffer

    return status, stdout, stderr


class SshBase(ICommandRunner):
    """
        The :class:`SshBase` object provides for the sharing of state and fuctional patterns
        for APIs between the :class:`SshSession` object and the :class:`SshAgent`.
    """
    def __init__(self, host: str, primary_credential: SshCredential, users: Optional[dict] = None,
                 port: int = 22, pty_params: Optional[dict] = None, called_id: Optional[str]=None,
                 aspects: AspectsCmd = DEFAULT_CMD_ASPECTS):

        self._host = host

        self._primary_credential = primary_credential
        self._username = primary_credential.username
        self._password = primary_credential.password
        self._keyfile = primary_credential.keyfile
        self._keypasswd = primary_credential.keypasswd
        self._allow_agent = primary_credential.allow_agent
        self._primitive = primary_credential.primitive

        self._users = users
        self._port = port
        self._pty_params = pty_params
        self._called_id = called_id
        self._target_tag = "[{}]".format(called_id)
        self._aspects = aspects

        self._ipaddr = socket.gethostbyname(self._host)

        self._user_lookup_table = {}
        self._group_lookup_table = {}

        if self._password is None and self._keyfile is None and not self._allow_agent:
            raise AKitInvalidConfigError("SshAgent requires either a 'password', an identity 'keyfile' or allow_agent=True be specified.") from None
        return

    @property
    def aspects(self) -> AspectsCmd:
        """
            The logging, iteration and other aspects that have been assigned to be used with interacting with the remote SSH service.
        """
        return self._aspects

    @property
    def host(self) -> str:
        """
            The target SSH host machine.
        """
        return self._host

    @property
    def ipaddr(self) -> str:
        """
            The IP address that was found to be associated with the specified host machine.
        """
        return self._ipaddr

    @property
    def port(self) -> int:
        """
            The port to use for SSH communications.  The default is 22.
        """
        return self._port

    @property
    def primary_credential(self):
        """
            Returns the 'primary' user credential provided for this ssh client object.
        """
        return self._primary_credential

    @property
    def pty_params(self) -> dict:
        """
            A dictionary of parameters that are to be passed when getting a PTY when interacting with the remote SSH server.
        """
        return self._pty_params

    @property
    def users(self) -> dict:
        """
            A dictionary of roles and user credentials that can be used for interaction with the remote SSH server.
        """
        return self._users

    def lookup_user_by_uid(self, uid: int, update: bool = True) -> Union[dict, None]:
        """
            A helper method used to lookup cached user information by uid.

            :param uid: The uid of a user.
            :param update: A boolean indicating that the table should be updated if the uid is not found.
        """
        username = None

        if uid in self._user_lookup_table:
            username = self._user_lookup_table[uid]
        elif update:
            status, stdout, _ = self.run_cmd("id -n -u %d" % uid)
            if status == 0:
                username = stdout.strip()
                self._user_lookup_table[uid] = username

        return username

    def lookup_group_by_uid(self, gid: int, update: bool = True) -> Union[dict, None]:
        """
            A helper method used to lookup cached group information by uid.

            :param gid: The gid of a group.
            :param update: A boolean indicating that the table should be updated if the uid is not found.

            :returns: The name of the group associated with the group id or None.
        """
        grpname = None

        if gid in self._group_lookup_table:
            grpname = self._group_lookup_table[gid]
        elif update:
            status, stdout, _ = self.run_cmd("id -n -g %d" % gid)
            if status == 0:
                grpname = stdout.strip()
                self._group_lookup_table[gid] = grpname

        return grpname

    def reboot(self, aspects: Optional[AspectsCmd] = None):
        """
            Reboots the designated host by running the 'reboot' command on the host.
            
            ..note: This method is not a typical SSH functionality, but it was added so that
                    automatic reconnects can be implemented as required on session objects.

        """
        raise AKitNotOverloadedError("SshBase.reboot must be overloaded by derived class '%s'." % type(self).__name__) from None
        reboot_cmd = 'reboot'
        try:
            status, stdout, stderr = self.run_cmd(reboot_cmd, aspects=aspects)
        except:
            # We might see a connection dropout exception here, so we need
            # to handle it.
            pass

        # Try to reconnect to the remote machine after the reboot.  We raise a AKitTimeoutError
        # if we do not reconnect before the reboot_timeout runs out.
        
        return

    def run_cmd(self, command: str, exp_status: Union[int, Sequence]=0, user: str = None, pty_params: dict = None, aspects: Optional[AspectsCmd] = None) -> Tuple[int, str, str]:
        """
            Runs a command on the designated host using the specified parameters.

            :param command: The command to run.
            :param exp_status: An integer or sequence of integers that specify the set of expected status codes from the command.
            :param user: The registered name of the user role to use to lookup the credentials for running the command.
            :param pty_params: The pty parameters to use to request a PTY when running the command.
            :param aspects: The run aspects to use when running the command.

            :returns: The status, stderr and stdout from the command that was run.
        """
        raise AKitNotOverloadedError("SshBase.run_cmd must be overloaded by derived class '%s'." % type(self).__name__) from None

    def verify_connectivity(self) -> bool:
        """
            Method that can be used to verify connectivity to the target computer.

            :returns: A boolean value indicating whether connectivity with the remote machine was successful.
        """
        vok = False

        hello_cmd = "echo Hello"
        status, stdout, _ = self.run_cmd(hello_cmd)
        if status == 0 and stdout.strip() == "Hello":
            vok = True

        return vok

    def _create_client(self, session_user: Optional[str] = None) -> paramiko.SSHClient:
        """
            Create an SSHClient to use for running commands and performing FTP operations.

            :param session_user: The user role and associated credentials to use when creating the SSHClient.

            :returns: An SSHClient object connected to the remote machine under the default or specified user credential.
        """
        cl_username = self._username
        cl_password = self._password
        cl_keyfile = self._keyfile
        cl_keypasswd = self._keypasswd
        cl_allow_agent = self._allow_agent

        if session_user is not None:
            if session_user not in self._users:
                errmsg_list = [
                    "No credentials found for the specified user '%s'." % session_user
                ]
                errmsg = os.linesep.join(errmsg_list)
                raise AKitInvalidConfigError(errmsg) from None

            user_creds = self._users[session_user]

            cl_username = user_creds["username"]
            cl_password = user_creds["password"]
            cl_keyfile = user_creds["keyfile"]
            cl_keypasswd = user_creds["keypasswd"]
            cl_allow_agent = user_creds["allow_agent"]

        pkey = None
        if cl_keyfile is not None:
            cl_keyfile = os.path.expanduser(os.path.expandvars(cl_keyfile))
            pkey = paramiko.rsakey.RSAKey.from_private_key_file(cl_keyfile, password=cl_keypasswd)

        ssh_client = paramiko.SSHClient()
        ssh_client.set_missing_host_key_policy(paramiko.AutoAddPolicy())
        ssh_client.connect(self._ipaddr, port=self._port, username=cl_username, password=cl_password, pkey=pkey, allow_agent=cl_allow_agent)
        return ssh_client

    def _directory(self, ssh_client: paramiko.SSHClient, root_dir: str) -> dict:
        """
            Private method that creates a directory listing for the folder.

            :param ssh_client: The SSHClient to use to for geting a list tree.
            :param root_dir: The directory to scan when creating the tree.

            :returns: A dictionary that contains information about the items in the target directory.
        """
        dir_info = {}

        if not self._primitive:
            transport = ssh_client.get_transport()
            try:

                sftp = paramiko.SFTPClient.from_transport(transport)
                try:
                    dir_info = sftp_list_directory(sftp, root_dir, self.lookup_user_by_uid, self.lookup_group_by_uid)
                finally:
                    sftp.close()
            finally:
                transport.close()
        else:
            dir_info = primitive_list_directory(ssh_client, root_dir)

        return dir_info

    def _directory_exists(self, ssh_client: paramiko.SSHClient, remotedir: str) -> bool:
        """
            Private helper method used to check if the remote directory exists.
        """
        exists = False

        if not self._primitive:
            transport = ssh_client.get_transport()
            try:
                sftp = paramiko.SFTPClient.from_transport(transport)
                try:
                    fsresult = sftp.stat(remotedir)
                    if fsresult.st_mode | stat.S_IFDIR:
                        exists = True
                finally:
                    sftp.close()
            finally:
                transport.close()
        else:
            primitive_directory_exists(ssh_client, remotedir)

        return exists

    def _directory_tree(self, ssh_client: paramiko.SSHClient, root_dir: str, depth: int = 1) -> dict:
        """
            Private method that creates a directory tree for the folder.

            :param ssh_client: The SSHClient to use to for geting a list tree.
            :param root_dir: The root directory to scan when creating the tree.
            :param depth: The dept to scan to

            :returns: A dictionary with a tree of information about the directory tree found on the remote system.
        """
        dir_info = {}

        if not self._primitive:
            transport = ssh_client.get_transport()
            try:

                sftp = paramiko.SFTPClient.from_transport(transport)
                try:
                    dir_info = sftp_list_tree(sftp, root_dir, self.lookup_user_by_uid, self.lookup_group_by_uid, max_depth=depth)
                finally:
                    sftp.close()
            finally:
                transport.close()
        else:
            dir_info = primitive_list_tree(ssh_client, root_dir, max_depth=depth)

        return dir_info

    def _log_command_result(self, command: str, status: int, stdout: str, stderr: str, exp_status: Union[int, Sequence[int]], logging_pattern: LoggingPattern):
        """
            Private method that handles the logging of command results based on the expected status and logging pattern
            specified.

            :param command: The command that was run.
            :param status: The status code returned from running the command.
            :param stdout: The contents of the standard output from running the command.
            :param stderr: The contents of the error output from running the command.
            :param exp_status: The status code or possible status codes that was expected to be returned from running the command.
            :param logging_pattern: The result logging pattern (LoggingPattern) to use when logging.
        """
        # pylint: disable=no-self-use

        tgt_tag = ""
        if self._target_tag is not None:
            tgt_tag = self._target_tag

        if isinstance(exp_status, int):
            if status == exp_status:
                if logging_pattern == LoggingPattern.ALL_RESULTS or logging_pattern == LoggingPattern.SUCCESS_ONLY:
                    logger.info(TEMPLATE_COMMAND_SUCCESS % (tgt_tag, command, stdout, stderr) )
        else:
            if status in exp_status:
                if logging_pattern == LoggingPattern.ALL_RESULTS or logging_pattern == LoggingPattern.SUCCESS_ONLY:
                    logger.info(TEMPLATE_COMMAND_SUCCESS % (tgt_tag, command, stdout, stderr))
            else:
                if logging_pattern == LoggingPattern.ALL_RESULTS or logging_pattern == LoggingPattern.FAILURE_ONLY:
                    logger.error(TEMPLATE_COMMAND_FAILURE % (tgt_tag, command, stdout, stderr))

        return

    def _run_cmd(self, ssh_runner: Union[paramiko.SSHClient, paramiko.Channel], command: str, exp_status: Union[int, Sequence]=0, user: str = None, pty_params: dict = None, aspects: Optional[AspectsCmd] = None) -> Tuple[int, str, str]:
        """
            Private method that handles the running commands in a generic way so the logic and run pattern code can be shared between agent and session objects.

            :param ssh_runner: A :class:`paramiko.SSHClient` or :class:`paramiko.Channel` object that can be used to run a command on a remote machine.
            :param command: The command that is to be run.
            :param exp_status: The status code or a sequence of possible status codes that are expected to be returned from running the command.
            :param user: The role name of the user account to utilize when looking up the credentials used to connect to the remote host.
            :param pty_params: A dictionary of parameters that are passed to paramiko to get a pty when running commands.
            :param aspects: The run aspects to use when running the command.

            :returns: The status, stderr and stdout from the command that was run.
        """
        # Go through the overrides and if they are not passed use the agent defaults.
        if pty_params is None:
            pty_params = self._pty_params

        if aspects is None:
            aspects = self._aspects

        if user is not None:
            if self._users is None:
                errmsg = "In order to pass a 'user' parameter, you must create the SSHAgent with a 'users' parameter with a dictionary of user credentials."
                raise AKitInvalidConfigError(errmsg) from None
            if user not in self._users:
                errmsg = "The specified 'user=%s' was not found in the users credentials provided to SSHAgent."

        completion_toctx = TimeoutContext(aspects.completion_timeout, aspects.completion_interval)
        
        inactivity_timeout = aspects.inactivity_timeout
        inactivity_interval = aspects.inactivity_interval
        monitor_delay = aspects.monitor_delay
        logging_pattern = aspects.logging_pattern

        status, stdout, stderr = None, None, None

        this_thr = threading.current_thread()
        monmsg= "Thread failed to exit monitored scope. thid=%s thname=%s cmd=%s" % (this_thr.ident, this_thr.name, command)

        # Run the command using the SINGULAR pattern, we run it once and then return the result
        if aspects.action_pattern == ActionPattern.SINGLE_CALL:

            # Setup a monitored scope for the call to the remote device in case of timeout failure
            with MonitoredScope("RUNCMD-SINGULAR", monmsg, completion_toctx, notify_delay=monitor_delay) as _:

                status, stdout, stderr = self._ssh_execute_command(ssh_runner, command, pty_params=pty_params,
                    inactivity_timeout=inactivity_timeout, inactivity_interval=inactivity_interval)

            self._log_command_result(command, status, stdout, stderr, exp_status, logging_pattern)

        # DO_UNTIL_SUCCESS, run the command until we get a successful expected result or a completion timeout has occured
        elif aspects.action_pattern == ActionPattern.DO_UNTIL_SUCCESS:

            completion_toctx.mark_begin()

            while True:

                # Setup a monitored scope for the call to the remote device in case of timeout failure
                with MonitoredScope("RUNCMD-DO_UNTIL_SUCCESS", monmsg, completion_toctx, notify_delay=monitor_delay) as _:
                    status, stdout, stderr = self._ssh_execute_command(ssh_runner, command, pty_params=pty_params,
                        inactivity_timeout=inactivity_timeout, inactivity_interval=inactivity_interval)

                self._log_command_result(command, status, stdout, stderr, exp_status, logging_pattern)

                if isinstance(exp_status, int):
                    if status == exp_status:
                        break
                elif status in exp_status:
                    break

                if completion_toctx.final_attempt:
                    what_for="command success"
                    details = [
                        "CMD: %s" % command,
                        "STDOUT:",
                        indent_lines(stdout, 1),
                        "STDERR:",
                        indent_lines(stderr, 1),
                    ]
                    toerr = completion_toctx.create_timeout(what_for=what_for, detail=details)
                    raise toerr

                elif not completion_toctx.should_continue():
                    completion_toctx.mark_final_attempt()

                time.sleep(completion_toctx.interval)

        # DO_WHILE_SUCCESS, run the command while it is succeeds or a completion timeout has occured
        elif aspects.action_pattern == ActionPattern.DO_WHILE_SUCCESS:

            completion_toctx.mark_begin()

            while True:

                with MonitoredScope("RUNCMD-DO_WHILE_SUCCESS", monmsg, completion_toctx, notify_delay=monitor_delay) as _:
                    status, stdout, stderr = self._ssh_execute_command(ssh_runner, command, pty_params=pty_params,
                        inactivity_timeout=inactivity_timeout, inactivity_interval=inactivity_interval)

                self._log_command_result(command, status, stdout, stderr, exp_status, logging_pattern)

                if isinstance(exp_status, int):
                    if status != exp_status:
                        break
                elif status not in exp_status:
                    break

                if completion_toctx.final_attempt:
                    what_for="command failure"
                    details = [
                        "CMD: %s" % command,
                        "STDOUT:",
                        indent_lines(stdout, 1),
                        "STDERR:",
                        indent_lines(stderr, 1),
                    ]
                    toerr = completion_toctx.create_timeout(what_for=what_for, detail=details)
                    raise toerr

                elif not completion_toctx.should_continue():
                    completion_toctx.mark_final_attempt()

                time.sleep(completion_toctx.interval)

        elif aspects.action_pattern == ActionPattern.SINGLE_CONNECTED_CALL or aspects.action_pattern == ActionPattern.DO_UNTIL_CONNECTION_FAILURE:
            errmsg = "SshAgent and SshSession currently do not support the SINGLE_CONNECTED_CALL or DO_UNTIL_CONNECTION_FAILURE action patterns."
            raise AKitSemanticError(errmsg) from None
        else:
            errmsg = "SshBase: Unknown ActionPattern encountered. action_pattern={}".format(aspects.action_pattern)
            raise AKitSemanticError(errmsg) from None

        return status, stdout, stderr

    def _file_exists(self, ssh_client: paramiko.SSHClient, remotepath: str) -> bool:
        """
            Private helper method used to check if the remote file exists.
        """
        exists = False

        if not self._primitive:
            transport = ssh_client.get_transport()
            try:
                sftp = paramiko.SFTPClient.from_transport(transport)
                try:
                    fsresult = sftp.stat(remotepath)
                    if not (fsresult.st_mode | stat.S_IFDIR):
                        exists = True
                finally:
                    sftp.close()
            finally:
                transport.close()
        else:
            exists = primitive_file_exists(ssh_client, remotepath)

        return exists

    def _file_pull(self, ssh_client: paramiko.SSHClient, remotepath: str, localpath: str):
        """
            Private helper method used to pull a remote file to a local file path.
        """
        if not self._primitive:
            transport = ssh_client.get_transport()
            try:
                sftp = paramiko.SFTPClient.from_transport(transport)
                try:
                    sftp.get(remotepath, localpath)
                finally:
                    sftp.close()
            finally:
                transport.close()
        else:
            primitive_file_pull(ssh_client, remotepath, localpath)

        return

    def _file_push(self, ssh_client: paramiko.SSHClient, localpath: str, remotepath: str):
        """
            Private helper method used to push a local file to a remote file path.
        """
        if not self._primitive:
            transport = ssh_client.get_transport()
            try:
                sftp = paramiko.SFTPClient.from_transport(transport)
                try:
                    sftp.put(localpath, remotepath)
                finally:
                    sftp.close()
            finally:
                transport.close()
        else:
            primitive_file_push(ssh_client, localpath, remotepath)

        return

    def _ssh_execute_command(self, ssh_runner, command: str, pty_params=None, inactivity_timeout: float=DEFAULT_SSH_TIMEOUT, inactivity_interval: float=DEFAULT_SSH_RETRY_INTERVAL, chunk_size: int=1024) -> Tuple[int, str, str]: # pylint: disable=no-self-use
        """
            Private helper method used to route the command running parameters to the correct routine based on whether the command is being run interactively
            through a session or is a single run command from an SshAgent.

            :param ssh_runner: A :class:`paramiko.SSHClient` or :class:`paramiko.Channel` object that can be used to run a command on a remote machine.
            :param command: The command that is to be run.
            :param pty_params: A dictionary of parameters that are passed to paramiko to get a pty when running commands.
            :param inactivity_timeout: A timeout for inactivity between the local machine and remote machine.
            :param inactivity_interval: The interval to wait between attempts to interact or read from the remote machine.
            :param chunk_size: The size of the buffer to use when reading results from the remote machine.
        """
        status, stdout, stderr = ssh_execute_command(ssh_runner, command, pty_params=pty_params, inactivity_timeout=inactivity_timeout, inactivity_interval=inactivity_interval, chunk_size=chunk_size)
        return status, stdout, stderr


class SshSession(SshBase):
    """
        The :class:`SshSession` provides an extension of the :class:`SshBase` class interface and api(s).  The :class:`SshSession` serves to
        act as a session scoping object that establishes a set of SSH credentials, settings and patterns that are to be used for a series of
        SSH operations.  The :class:`SshSession` also holds open the SSHClient for its entire scope and ensures the SSHClient is closed and
        cleaned up properly when the :class:`SshSession` goes out of scope.
    """
    def __init__(self, host: str, primary_credential: SshCredential, users: Optional[dict] = None, port:int=22,
                 pty_params: Optional[dict] = None, session_user=None, interactive=False, basis_session: Optional["SshSession"]=None,
                 called_id: Optional[str]=None, aspects: AspectsCmd=DEFAULT_CMD_ASPECTS):
        SshBase.__init__(self, host, primary_credential, users=users, port=port, pty_params=pty_params, called_id=called_id, aspects=aspects)

        self._session_user = session_user
        self._interactive = interactive
        self._read_timeout = 2

        self._ssh_client = None
        self._ssh_runner = None
        self._basis_session = basis_session
        return

    def __enter__(self):
        """
            Method that implements the __enter__ symantics used for 'with' statements.
        """
        self._ssh_client = self._create_client()

        if self._interactive:
            inactivity_timeout = self._aspects.inactivity_timeout

            channel = self._ssh_client.get_transport().open_session(timeout=inactivity_timeout)

            if self._pty_params is not None:
                channel.get_pty(**self._pty_params)
            else:
                channel.get_pty()

            channel.invoke_shell()
            channel.sendall("export PS1=%s\n" % INTERACTIVE_PROMPT)

            leader = b""
            try:
                while True:
                    out = channel.in_buffer.read(1028, self._read_timeout)
                    leader += out
            except paramiko.buffered_pipe.PipeTimeout as _:
                # We don't need to do anything if the pipe, times out.
                # it means we are likely at the command prompt and that
                # there is nothing to read.
                pass

            self._ssh_runner = channel
        else:
            self._ssh_runner = self._ssh_client

        return self

    def __exit__(self, ex_type: type, ex_val: Exception, ex_tb: TracebackType) -> bool:
        """
            Method that implements the __exit__ symantics used for 'with' statements.

            :returns: Returns false indicating that exceptions are not handled.
        """
        self.close()
        handled = False
        return handled

    def close(self):
        """
            Closes the SSH session and the assocatied SSH connection.
        """
        # Only close the client if this session owns the client
        if self._basis_session is None:
            self._ssh_client.close()
        return

    def reboot(self, aspects: Optional[AspectsCmd] = None):
        """
            Reboots the designated host by running the 'reboot' command on the host.
            
            ..note: This method is not a typical SSH functionality, but it was added so that
                    automatic reconnects can be implemented as required on session objects.

        """
        reboot_cmd = 'reboot'
        try:
            status, stdout, stderr = self.run_cmd(reboot_cmd, aspects=aspects)
        except:
            # We might see a connection dropout exception here, so we need
            # to handle it.
            pass

        # Since we are a session, we need to reconnect to the remote machine after the reboot.
        # We raise a AKitTimeoutError if we do not reconnect before the reboot_timeout runs out.
        
        return

    def run_cmd(self, command: str, exp_status: Union[int, Sequence]=0, user: str = None, pty_params: dict = None, aspects: Optional[AspectsCmd] = None) -> Tuple[int, str, str]:
        """
            Runs a command on the designated host using the current session SSH session and client.

            :param command: The command to run.
            :param exp_status: An integer or sequence of integers that specify the set of expected status codes from the command.
            :param user: The registered name of the user role to use to lookup the credentials for running the command.
            :param pty_params: The pty parameters to use to request a PTY when running the command.
            :param aspects: The run aspects to use when running the command.

            :returns: The status, stderr and stdout from the command that was run.
        """

        status, stdout, stderr = self._run_cmd(self._ssh_runner, command, user=user, pty_params=pty_params, aspects=aspects)

        return status, stdout, stderr

    def directory(self, root_dir: str) -> dict:
        """
            Method that creates a directory listing for the folder.

            :param root_dir: The directory to scan when creating the tree.

            :returns: A dictionary that contains information about the items in the target directory.
        """
        dir_info = self._directory(self._ssh_client, root_dir)

        return dir_info

    def directory_exists(self, remotedir: str) -> bool:
        """
            Method used to pull a remote directory to check for existance.

            :param remotedir: The remote directory path to check for existance.

            :returns: A boolean value indicating if the remote file exists.
        """
        exists = self._file_pull(self._ssh_client, remotedir)

        return exists

    def directory_tree(self, root_dir: str, depth: int = 1) -> dict:
        """
            Method that creates a directory tree for the folder.

            :param root_dir: The root directory to scan when creating the tree.
            :param depth: The dept to scan to

            :returns: A dictionary with a tree of information about the directory tree found on the remote system.
        """
        dir_info = self._directory_tree(self._ssh_client, root_dir, depth=depth)

        return dir_info

    def file_exists(self, remotepath: str) -> bool:
        """
            Method used to pull a remote file to check for existance.

            :param remotepath: The remote file path to check for existance.

            :returns: A boolean value indicating if the remote file exists.
        """
        exists = self._file_pull(self._ssh_client, remotepath)

        return exists

    def file_pull(self, remotepath: str, localpath: str):
        """
            Method used to pull a remote file to a local file path.

            :param remotepath: The remote file path to pull to the local file.
            :param localpath: The local file path to pull the content to.
        """
        self._file_pull(self._ssh_client, remotepath, localpath)

        return

    def file_push(self, localpath: str, remotepath: str):
        """
            Method used to push a local file to a remote file path.

            :param localpath: The local file path to push the content of to the remote file.
            :param remotepath: The remote file path to push content to.
        """
        self._file_push(self._ssh_client, localpath, remotepath)

        return

    def open_session(self, primitive: bool = False, pty_params: Optional[dict] = None, interactive=False, basis_session: Optional["SshSession"] = None,
                     aspects: Optional[AspectsCmd] = None) -> "SshSession":
        """
            Provies a mechanism to create a :class:`SshSession` object with derived settings.  This method allows various parameters for the session
            to be overridden.  This allows for the performing of a series of SSH operations under a particular set of shared settings and or credentials.

            :param primitive: Use primitive mode for FTP operations for the session.
            :param pty_params: The default pty parameters to use to request a PTY when running commands through the session.
            :param interactive: Creates an interactive session which holds open an interactive shell so commands can interact in the shell.
            :param basis_session: An optional SshSession instance to use as a session basis.  This allows re-use of sessions.
            :param aspects: The default run aspects to use for the operations performed by the session.
        """

        if aspects is None:
            aspects = self._aspects

        session = None
        if basis_session is not None:
            bs = basis_session
            session = SshSession(bs._host, bs._primary_credential, users=bs._users, port=bs._port, pty_params=pty_params,
                             interactive=interactive, basis_session=basis_session, aspects=aspects)
        else:
            session = SshSession(self._host, self._primary_credential, users=self._users, port=self._port, pty_params=pty_params,
                             interactive=interactive, basis_session=basis_session, aspects=aspects)
        return session

    def _create_client(self, session_user: Optional[str] = None) -> paramiko.SSHClient:
        """
            Create an SSHClient to use for running commands and performing FTP operations.

            :param session_user: The user role and associated credentials to use when creating the SSHClient.

            :returns: An SSHClient object connected to the remote machine under the default or specified user credential.
        """
        ssh_client = None
        if self._basis_session is not None:
            ssh_client = self._basis_session._ssh_client
        else:
            ssh_client = SshBase._create_client(self, session_user=session_user)
        return ssh_client

    def _ssh_execute_command(self, ssh_runner, command: str, pty_params=None, inactivity_timeout: float=DEFAULT_SSH_TIMEOUT, inactivity_interval: float=DEFAULT_SSH_RETRY_INTERVAL, chunk_size: int=1024) -> Tuple[int, str, str]:
        """
            Private helper method used to route the command running parameters to the correct routine based on whether the command is being run interactively
            through a session or is a single run command from an SshAgent.

            :param ssh_runner: A :class:`paramiko.SSHClient` or :class:`paramiko.Channel` object that can be used to run a command on a remote machine.
            :param command: The command that is to be run.
            :param pty_params: A dictionary of parameters that are passed to paramiko to get a pty when running commands.
            :param inactivity_timeout: A timeout for inactivity between the local machine and remote machine.
            :param inactivity_interval: The interval to wait between attempts to interact or read from the remote machine.
            :param chunk_size: The size of the buffer to use when reading results from the remote machine.
        """
        if self._interactive:
            status, stdout, stderr = ssh_execute_command_in_channel(ssh_runner, command, read_timeout=self._read_timeout, inactivity_timeout=inactivity_timeout, inactivity_interval=inactivity_interval)
        else:
            status, stdout, stderr = ssh_execute_command(ssh_runner, command, pty_params=pty_params, inactivity_timeout=inactivity_timeout, inactivity_interval=inactivity_interval)
        return status, stdout, stderr

class SshAgent(SshBase, LandscapeDeviceExtension):
    """
        The :class:`SshAgent` provides an extension of the :class:`SshBase` class interface and api(s).  The :class:`SshAgent` manages connection
        and interop settings that are used to perform operations and interact with a remote SSH service.  The :class:`SshAgent` provides standard
        APIs for running command and transferring files along with code that helps ensure commands are logged cleanly.  The :class:`SshAgent` also
        provides run patterning to help eliminate duplication of code associated with running SSH commands in loops.
    """
    def __init__(self, host: str, primary_credential: SshCredential, users: Optional[dict] = None, port: int = 22,
                 pty_params: Optional[dict] = None, called_id: Optional[str]=None, aspects: AspectsCmd = DEFAULT_CMD_ASPECTS):
        SshBase.__init__(self, host, primary_credential, users=users, port=port, pty_params=pty_params, called_id=called_id, aspects=aspects)
        LandscapeDeviceExtension.__init__(self)
        return

    def initialize(self, coord_ref: weakref.ReferenceType, basedevice_ref: weakref.ReferenceType, extid: str, location: str, configinfo: dict):
        """
            Initializes the landscape device extension. It is not required to call this function in order to use an SSHAgent, it is used
            when an agent is attached to a LandscapeDevice as a device extension.

            :param coord_ref: A weak reference to the coordinator that is managing interactions through this
                              device extension.
            :param extid: A unique reference that can be used to identify this device via the coordinator even if its location changes.
            :param location: The location reference where this device can be found via the coordinator.
            :param configinfo: The configuration information dictionary from the landscape file for the specified host.
        """
        LandscapeDeviceExtension.initialize(self, coord_ref, basedevice_ref, extid, location, configinfo)
        return

    def open_session(self, primitive: bool = False, pty_params: Optional[dict] = None, interactive=False, basis_session: Optional[SshSession] = None,
                     aspects: Optional[AspectsCmd] = None) -> SshSession:
        """
            Provies a mechanism to create a :class:`SshSession` object with derived settings.  This method allows various parameters for the session
            to be overridden.  This allows for the performing of a series of SSH operations under a particular set of shared settings and or credentials.

            :param primitive: Use primitive mode for FTP operations for the session.
            :param pty_params: The default pty parameters to use to request a PTY when running commands through the session.
            :param interactive: Creates an interactive session which holds open an interactive shell so commands can interact in the shell.
            :param basis_session: An optional SshSession instance to use as a session basis.  This allows re-use of sessions.
            :param aspects: The default run aspects to use for the operations performed by the session.
        """

        if aspects is None:
            aspects = self._aspects

        session = None
        if basis_session is not None:
            bs = basis_session
            session = SshSession(bs._host, bs._primary_credential, users=bs._users, port=bs._port, pty_params=pty_params,
                             interactive=interactive, basis_session=basis_session, aspects=aspects)
        else:
            session = SshSession(self._host, self._primary_credential, users=self._users, port=self._port, pty_params=pty_params,
                             interactive=interactive, basis_session=basis_session, aspects=aspects)
        return session

    def reboot(self, aspects: Optional[AspectsCmd] = None):
        """
            Reboots the designated host by running the 'reboot' command on the host.
            
            ..note: This method is not a typical SSH functionality, but it was added so that
                    automatic reconnects can be implemented as required on session objects.

        """
        reboot_cmd = 'reboot'
        try:
            status, stdout, stderr = self.run_cmd(reboot_cmd, aspects=aspects)
        except:
            # We might see a connection dropout exception here, so we need
            # to handle it.
            pass

        # Since we are NOT a session, we need to verify the remote machine is back up by attempting to
        # connect to the machine and run a command. We raise a AKitTimeoutError if we cannot connect and
        # run a command before the reboot_timeout runs out.
        
        return

    def run_cmd(self, command: str, exp_status: Union[int, Sequence]=0, user: str = None, pty_params: dict = None, aspects: Optional[AspectsCmd] = None, ssh_client: Optional[paramiko.SSHClient]=None) -> Tuple[int, str, str]: # pylint: disable=arguments-differ
        """
            Runs a command on the designated host using the specified parameters.

            :param command: The command to run.
            :param exp_status: An integer or sequence of integers that specify the set of expected status codes from the command.
            :param user: The registered name of the user role to use to lookup the credentials for running the command.
            :param pty_params: The pty parameters to use to request a PTY when running the command.
            :param aspects: The run aspects to use when running the command.
            :param ssh_client: An optional connected SSHClient that should be for the operation.

            :returns: The status, stderr and stdout from the command that was run.
        """
        cleanup_client = False

        status, stdout, stderr = None, None, None

        try:
            if ssh_client is None:
                # If we are not passed in a client to use, we create one
                # and then clean it up before returning
                ssh_client = self._create_client()
                cleanup_client = True

            status, stdout, stderr = self._run_cmd(ssh_client, command, exp_status=exp_status, user=user, pty_params=pty_params, aspects=aspects)

        finally:
            # If we created a client, clean it up
            if cleanup_client:
                ssh_client.close()
                del ssh_client

        return status, stdout, stderr

    def directory_tree(self, root_dir: str, depth: int = 1, ssh_client: Optional[paramiko.SSHClient] = None) -> dict:
        """
            Method that creates a directory tree for the folder.

            :param root_dir: The root directory to scan when creating the tree.
            :param depth: The dept to scan to
            :param ssh_client: An optional connected SSHClient that should be for the operation.

            :returns: A dictionary with a tree of information about the directory tree found on the remote system.
        """
        dir_info = {}

        cleanup_client = False
        try:
            if ssh_client is None:
                ssh_client = self._create_client()
                cleanup_client = True

            dir_info = self._directory_tree(ssh_client, root_dir, depth=depth)

        finally:
            if cleanup_client:
                ssh_client.close()
                del ssh_client

        return dir_info

    def directory(self, root_dir: str, ssh_client: Optional[paramiko.SSHClient] = None) -> dict:
        """
            Method that creates a directory listing for the folder.

            :param root_dir: The directory to scan when creating the tree.
            :param ssh_client: An optional connected SSHClient that should be for the operation.

            :returns: A dictionary that contains information about the items in the target directory.
        """
        dir_info = {}

        cleanup_client = False
        try:
            if ssh_client is None:
                ssh_client = self._create_client()
                cleanup_client = True

            dir_info = self._directory(ssh_client, root_dir)

        finally:
            if cleanup_client:
                ssh_client.close()
                del ssh_client

        return dir_info

    def directory_exists(self, remotedir: str, ssh_client: Optional[paramiko.SSHClient] = None) -> bool:
        """
            Method used to check if a remote directory exists.

            :param remotedir: The remote remotedir path to check for existance.
            :param ssh_client: An optional connected SSHClient that should be for the operation.
        """
        exists = False

        try:
            if ssh_client is None:
                ssh_client = self._create_client()
                cleanup_client = True

            self._directory_exists(ssh_client, remotedir)

        finally:
            if cleanup_client:
                ssh_client.close()
                del ssh_client

        return exists

    def file_exists(self, remotepath: str, ssh_client: Optional[paramiko.SSHClient] = None) -> bool:
        """
            Method used to check if a remote file exists.

            :param remotepath: The remote file path to check for existance.
            :param ssh_client: An optional connected SSHClient that should be for the operation.
        """
        exists = False

        try:
            if ssh_client is None:
                ssh_client = self._create_client()
                cleanup_client = True

            exists = self._file_exists(ssh_client, remotepath)

        finally:
            if cleanup_client:
                ssh_client.close()
                del ssh_client

        return exists

    def file_pull(self, remotepath: str, localpath: str, ssh_client: Optional[paramiko.SSHClient] = None):
        """
            Method used to pull a remote file to a local file path.

            :param remotepath: The remote file path to pull to the local file.
            :param localpath: The local file path to pull the content to.
            :param ssh_client: An optional connected SSHClient that should be for the operation.
        """
        cleanup_client = False
        try:
            if ssh_client is None:
                ssh_client = self._create_client()
                cleanup_client = True

            self._file_pull(ssh_client, remotepath, localpath)

        finally:
            if cleanup_client:
                ssh_client.close()
                del ssh_client

        return

    def file_push(self, localpath: str, remotepath: str, ssh_client: Optional[paramiko.SSHClient] = None):
        """
            Method used to push a local file to a remote file path.

            :param localpath: The local file path to push the content of to the remote file.
            :param remotepath: The remote file path to push content to.
            :param ssh_client: An optional connected SSHClient that should be for the operation.
        """
        cleanup_client = False
        try:
            if ssh_client is None:
                ssh_client = self._create_client()
                cleanup_client = True

            self._file_push(ssh_client, localpath, remotepath)

        finally:
            if cleanup_client:
                ssh_client.close()
                del ssh_client

        return




__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os

import click

from akit.environment.variables import LOG_LEVEL_NAMES

@click.group("upnp")
def group_akit_generators_upnp():
    return

@click.command("generate")
def command_akit_generators_upnp_generate():
    # We want to run the scan as a console application so we import the console
    # environment in order to setup the appropriate logging
    import akit.activation.console

    from akit.interop.upnp.generator.upnpgenerator import generate_service_proxies

    from akit.interop.upnp.paths import DIR_UPNP_GENERATOR_STANDARD_SERVICES
    from akit.interop.upnp.paths import DIR_UPNP_EXTENSIONS_STANDARD_SERVICES

    from akit.interop.upnp.paths import DIR_UPNP_SCAN_INTEGRATION_SERVICES
    from akit.interop.upnp.paths import DIR_UPNP_EXTENSIONS_INTEGRATION_SERVICES

    if os.path.exists(DIR_UPNP_GENERATOR_STANDARD_SERVICES):
        generate_service_proxies(DIR_UPNP_GENERATOR_STANDARD_SERVICES, DIR_UPNP_EXTENSIONS_STANDARD_SERVICES)

    if os.path.exists(DIR_UPNP_SCAN_INTEGRATION_SERVICES):
        generate_service_proxies(DIR_UPNP_SCAN_INTEGRATION_SERVICES, DIR_UPNP_EXTENSIONS_INTEGRATION_SERVICES)

    return

@click.command("scan")
def command_akit_generators_upnp_scan():
    # We want to run the scan as a console application so we import the console
    # environment in order to setup the appropriate logging
    import akit.activation.console

    from akit.wellknown.singletons import LandscapeSingleton
    from akit.coupling.upnpcoordinatorintegration import UpnpCoordinatorIntegration

    # ==================== Landscape Initialization =====================
    # The first stage of standing up the test landscape is to create and
    # initialize the Landscape object.  If more than one thread calls the
    # constructor of the Landscape, object, the other thread will block
    # until the first called has initialized the Landscape and released
    # the gate blocking other callers.
    lscape = LandscapeSingleton()

    # Give the UpnpCoordinatorIntegration an opportunity to register itself, we are
    # doing this in this way to simulate test framework startup.
    UpnpCoordinatorIntegration.attach_to_framework(lscape)

    # Finalize the registration process and transition the landscape
    # to the activation phase
    lscape.activate_integration()

    # Give the UpnpCoordinatorIntegration an opportunity to attach to its
    # environment and determine if the resources requested and the
    # resource configuration match
    UpnpCoordinatorIntegration.attach_to_environment()

    # Finalize the activation process and transition the landscape
    # to fully active where all APIs are available.
    lscape.activate_operations(allow_missing_devices=True, allow_unknown_devices=True, upnp_recording=True)

    # Make initial contact with all of the devices, this will trigger a scan
    # and the scan will populate the appropriate xml metadata directories
    UpnpCoordinatorIntegration.establish_presence()

    return

group_akit_generators_upnp.add_command(command_akit_generators_upnp_generate)
group_akit_generators_upnp.add_command(command_akit_generators_upnp_scan)



"""
.. module:: taskbase
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: A module that provides the base task object that other tasks derive
               from.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Optional

from akit.exceptions import AKitNotOverloadedError

class TaskBase:

    def __init__(self, ordinal: str, label: str, task_info: dict, logger):
        self._ordinal = ordinal
        self._label = label
        self._task_info = task_info
        self._logger = logger

        self._onfailure = None
        if "onfailure" in task_info:
            self._onfailure = task_info["onfailure"]

        self._onsuccess = None
        if "onsuccess" in task_info:
            self._onsuccess = task_info["onsuccess"]

        return

    @property
    def label(self) -> str:
        return self._label

    @property
    def onfailure(self) -> str:
        return self._onfailure

    @property
    def onsuccess(self) -> str:
        return self._onsuccess

    @property
    def ordinal(self) -> str:
        return self._ordinal

    @property
    def task_info(self) -> dict:
        return self._task_info

    def execute(self, parameters: Optional[dict]=None, topology: Optional[dict]=None, **kwargs) -> int:
        raise AKitNotOverloadedError("TaskBase->execute must be overloaded by derived classes.") from None


"""
.. module:: resources
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: A set of standardized decorations that are utilized to declare integrations
               scope and resource factory functions.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"


from typing import Callable, Dict, Optional, TypeVar

import inspect
import os

from akit.compat import NoneType

from akit.coupling.integrationcoupling import IntegrationCoupling
from akit.testing.testplus.scopecoupling import ScopeCoupling

from akit.exceptions import AKitSemanticError

from akit.testing.testplus.registration.resourceregistry import resource_registry

from akit.testing.testplus.registration.integrationsource import IntegrationSource
from akit.testing.testplus.registration.resourcesource import ResourceSource
from akit.testing.testplus.registration.scopesource import ScopeSource

_IntegrationSubscriberType = TypeVar("_IntegrationSubscriberType", bound=Callable[..., object])

def integration(*, constraints: Optional[Dict]=None):
    """
        The `integration` decorator is used to declare a function as the source
        of an integration resource.
    """
    def decorator(source_function: Callable) -> Callable:
        nonlocal constraints

        signature = inspect.signature(source_function)
        integration_context = signature.return_annotation

        resource_type = None

        if integration_context == inspect._empty:
            errmsg = "Parameters factories for 'integration' functions must have an annotated return type."
            raise AKitSemanticError(errmsg) from None
        else:
            if integration_context._name == "Generator":
                ra_yield_type, ra_send_type, ra_return_type = integration_context.__args__
                if ra_yield_type is not NoneType:
                    resource_type = ra_yield_type
                elif ra_return_type is not NoneType:
                    resource_type = ra_return_type
            elif issubclass(integration_context, IntegrationCoupling):
                raise AKitSemanticError("Your resource function is returning an integration instead of yielding one.") from None
            else:
                raise AKitSemanticError("You must pass a IntegrationCoupling derived object.") from None

        if resource_type is not None:

            if not issubclass(resource_type, IntegrationCoupling):
                raise AKitSemanticError("The 'integration' decorator can only be used on resources that inherit from the 'IntegrationCoupling'.") from None

            isource = IntegrationSource(source_function, resource_type, constraints)
            resource_registry.register_integration_source(isource)
        else:
            errmsg_lines = [
                "Unable to determine the resource type of the function which the 'integration' decorator was applied too.",
                "FUNCTION: {}".format(signature)
            ]
            errmsg = os.linesep.join(errmsg_lines)
            raise AKitSemanticError(errmsg) from None

        return source_function
    return decorator

def resource(*, query_function: Optional[Callable]=None, constraints: Optional[Dict]=None):
    def decorator(source_function: Callable) -> Callable:
        nonlocal constraints

        signature = inspect.signature(source_function)
        resource_context = signature.return_annotation

        resource_type = None

        if resource_context == inspect._empty:
            errmsg = "Parameters factories or 'resource' functions must have an annotated return type."
            raise AKitSemanticError(errmsg) from None
        elif hasattr(resource_context, "_name") and resource_context._name == "Generator":
            ra_yield_type, ra_send_type, ra_return_type = resource_context.__args__
            if ra_yield_type is not NoneType:
                resource_type = ra_yield_type
            elif ra_return_type is not NoneType:
                resource_type = ra_return_type
        elif issubclass(resource_context, IntegrationCoupling):
            resource_type = resource_context
        else:
            resource_type = resource_context
        
        if resource_type is not None:

            sref = ResourceSource(source_function, query_function, resource_type, constraints)
            resource_registry.register_resource_source(sref)
        else:
            errmsg_lines = [
                "Unable to determine the resource type of the function which the 'resource' decorator was applied too.",
                "FUNCTION: {}".format(signature)
            ]
            errmsg = os.linesep.join(errmsg_lines)
            raise AKitSemanticError(errmsg) from None

        return source_function
    return decorator

def scope(*, query_function: Optional[Callable]=None, constraints: Optional[Dict]=None):
    """
        The `scope` decorator is used to declare a function as the source
        of an scope resource.
    """
    def decorator(source_function: Callable) -> Callable:
        nonlocal constraints

        signature = inspect.signature(source_function)
        scope_context = signature.return_annotation

        resource_type = None

        if scope_context == inspect._empty:
            errmsg = "Parameters factories or 'scope' functions must have an annotated return type."
            raise AKitSemanticError(errmsg) from None
        elif scope_context._name == "Generator":
            ra_yield_type, ra_send_type, ra_return_type = scope_context.__args__
            if ra_yield_type is not NoneType:
                resource_type = ra_yield_type
            elif ra_return_type is not NoneType:
                resource_type = ra_return_type
        elif issubclass(scope_context, ScopeCoupling):
            resource_type = scope_context
        else:
            raise AKitSemanticError("You must pass a ScopeCoupling derived object.") from None
        
        if resource_type is not None:

            if not issubclass(resource_type, ScopeCoupling):
                raise AKitSemanticError("The 'scope' decorator can only be used on resources that inherit from the 'scopecoupling'.") from None

            ssource = ScopeSource(source_function, query_function, resource_type, constraints)
            resource_registry.register_scope_source(ssource)
        else:
            errmsg_lines = [
                "Unable to determine the resource type of the function which the 'scope' decorator was applied too.",
                "FUNCTION: {}".format(signature)
            ]
            errmsg = os.linesep.join(errmsg_lines)
            raise AKitSemanticError(errmsg) from None

        return source_function
    return decorator




#!/usr/bin/env python3
"""
.. module:: akitcommand
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: The script entrypoint for the 'akit' command.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import click

from akit.cli.cmdtree.databases import group_akit_databases
from akit.cli.cmdtree.generators import group_akit_generators
from akit.cli.cmdtree.landscape import group_akit_landscape
from akit.cli.cmdtree.network import group_akit_network
from akit.cli.cmdtree.workflow import group_akit_workflow
from akit.cli.cmdtree.servers import group_akit_servers
from akit.cli.cmdtree.testing import group_akit_testing
from akit.cli.cmdtree.utilities import group_akit_utilities

@click.group("akit")
@click.option('-v', '--verbose', count=True)
def akit_root_command(verbose):

    from akit.environment.variables import AKIT_VARIABLES

    if verbose == 0:
        AKIT_VARIABLES.AKIT_INTERACTIVE_CONSOLE = True
    else:
        AKIT_VARIABLES.AKIT_INTERACTIVE_CONSOLE = False

        if verbose == 1:
            AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE = "INFO"
        elif verbose == 2:
            AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE = "DEBUG"
        elif verbose > 2:
            AKIT_VARIABLES.AKIT_LOG_LEVEL_CONSOLE = "NOTSET"

    return

akit_root_command.add_command(group_akit_databases)
akit_root_command.add_command(group_akit_generators)
akit_root_command.add_command(group_akit_landscape)
akit_root_command.add_command(group_akit_network)
akit_root_command.add_command(group_akit_workflow)
akit_root_command.add_command(group_akit_servers)
akit_root_command.add_command(group_akit_testing)
akit_root_command.add_command(group_akit_utilities)

if __name__ == '__main__':
    akit_root_command()


"""
.. module:: upnprootdevice
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module containing the :class:`UpnpRootDevice` class and associated diagnostic.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>

"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from optparse import Option
from typing import Dict, List, Optional, Tuple, Union, TYPE_CHECKING

import os
import re
import threading
import traceback
import uuid
import weakref

from datetime import datetime, timedelta
from urllib.parse import urlparse

from xml.etree.ElementTree import tostring as xml_tostring
from xml.etree.ElementTree import fromstring as xml_fromstring
from xml.etree.ElementTree import Element, ElementTree
from xml.etree.ElementTree import register_namespace

import requests

from requests.compat import urljoin

from akit.exceptions import AKitCommunicationsProtocolError, AKitHTTPRequestError, AKitNotOverloadedError, AKitRuntimeError
from akit.extensible import generate_extension_key
from akit.interop.upnp.upnperrors import UpnpError
from akit.paths import normalize_name_for_path

from akit.interop.landscaping.landscapedeviceextension import LandscapeDeviceExtension

from akit.interop.upnp.aliases import StrSvcId, StrSvcType, StrSubId
from akit.interop.upnp.upnpconstants import TIMEDELTA_RENEWAL_WINDOW
from akit.interop.upnp.upnpprotocol import MSearchRouteKeys
from akit.interop.upnp.devices.upnpdevice import UpnpDevice
from akit.interop.upnp.devices.upnpembeddeddevice import UpnpEmbeddedDevice
from akit.interop.upnp.xml.upnpdevice1 import UPNP_DEVICE1_NAMESPACE, UpnpDevice1Device, UpnpDevice1SpecVersion
from akit.interop.upnp.soap import NS_UPNP_EVENT

from akit.interop.upnp.paths import DIR_UPNP_SCAN_INTEGRATION_EMBEDDEDDEVICES
from akit.interop.upnp.paths import DIR_UPNP_SCAN_INTEGRATION_ROOTDEVICES
from akit.interop.upnp.paths import DIR_UPNP_SCAN_INTEGRATION_SERVICES
from akit.interop.upnp.upnpprotocol import MSearchKeys

from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

from akit.interop import upnp as upnp_module

from akit.networking.constants import AKitHttpHeaders
from akit.raisefor import raise_for_http_status

# Types imported only for type checking purposes
if TYPE_CHECKING:
    from akit.interop.upnp.upnpfactory import UpnpFactory

UPNP_DIR = os.path.dirname(upnp_module.__file__)

REGEX_SUBSCRIPTION_TIMEOUT = re.compile("^Second-([0-9]+|infinite)", flags=re.IGNORECASE)

def device_description_load(location: str) -> Union[ElementTree, None]:
    """
        Gets the description document from the specified URL and loads the contents into an XML ElementTree.

        :param location: The url of where to get the device description.

        :returns: The XML ElementTree for the XML content from the device description or None.
    """
    docTree = None

    resp = requests.get(location)
    if resp.status_code == 200:
        xmlcontent = resp.content

        docTree = ElementTree(xml_fromstring(xmlcontent))

    return  docTree


def device_description_find_components(location: str, docTree: ElementTree, namespaces: dict = {"": UPNP_DEVICE1_NAMESPACE}) -> Tuple[Element, str, str, str, str, str]:
    """
        Processes the device description and find the device description components.  The device
        description components are the manufacturer, modelName, modelNumber, modelDescription
        and the device description node.

        :param location: The URL location of where the device description document was obtained.
        :param docTree: The XML document ElementTree object.
        :param namespaces: A dictionary of namespaces to use when processing the device description document.

        :returns: A tuple with the (device node, urlBase, manufacturer, modelName, modelNumber, modelDescription)
    """
    # pylint: disable=dangerous-default-value

    devNode = docTree.find("device", namespaces=namespaces)

    urlBase = None
    baseURLNode = devNode.find("URLBase", namespaces=namespaces)
    if baseURLNode is not None:
        urlBase = baseURLNode.text

    url_parts = urlparse(location)
    host = url_parts.netloc

    # If urlBase was not set we need to try to use the schema and host as the urlBase
    if urlBase is None:
        urlBase = "%s://%s" % (url_parts.scheme, host)

    manufacturer = None
    modelName = None
    modelNumber = None
    modelDescription = None

    manufacturerNode = devNode.find("manufacturer", namespaces=namespaces)
    if manufacturerNode is not None:
        manufacturer = manufacturerNode.text

    modelNameNode = devNode.find("modelName", namespaces=namespaces)
    if modelNameNode is not None:
        modelName = modelNameNode.text

    modelNumberNode = devNode.find("modelNumber", namespaces=namespaces)
    if modelNumberNode is not None:
        modelNumber = modelNumberNode.text

    modelDescNode = devNode.find("modelDescription", namespaces=namespaces)
    if modelDescNode is not None:
        modelDescription = modelDescNode.text

    return devNode, urlBase, manufacturer, modelName, modelNumber, modelDescription

class UpnpRootDevice(UpnpDevice, LandscapeDeviceExtension):
    """
        The UPNP Root device is the base device for the hierarchy that is
        associated with a unique network devices location.  The :class:`UpnpRootDevice`
        and its subdevices are linked by thier location url.

        http://www.upnp.org/specs/arch/UPnP-arch-DeviceArchitecture-v1.0.pdf
    """

    MANUFACTURER = "unknown"
    MODEL_NUMBER = "unknown"
    MODEL_DESCRIPTION = "unknown"

    SERVICE_NAMES = []

    def __init__(self, manufacturer: str, modelNumber: str, modelDescription):
        """
            Creates a root device object.
        """
        super(UpnpRootDevice, self).__init__()

        if self.MANUFACTURER != manufacturer:
            self.MANUFACTURER = manufacturer
        if self.MODEL_NUMBER != modelNumber:
            self.MODEL_NUMBER = modelNumber
        if self.MODEL_DESCRIPTION != modelDescription:
            self.MODEL_DESCRIPTION = modelDescription

        self._extra = {}
        self._cachecontrol = None
        self._ext = None
        self._server = None
        self._st = None
        self._usn = None
        self._usn_dev = None
        self._usn_cls = None

        self._specVersion = None

        self._host = None
        self._ip_address = None
        self._routes = None
        self._primary_route = None

        self._devices = {}
        self._device_descriptions = {}

        self._auto_subscribe = False
        self._mode = None

        self._root_device_lock = threading.RLock()

        self._subscriptions: Dict[StrSvcType, Tuple[StrSvcId, datetime]] = {}
        self._sid_to_service_lookup: Dict[str, UpnpServiceProxy] = {}
        self._variables = {}

        self._last_alive = datetime.min
        self._last_byebye = datetime.min
        self._available = False
        return

    @property
    def available(self) -> bool:
        """
            Returns the status of the device.
        """
        return self._available

    @property
    def cachecontrol(self) -> str:
        """
            The cache control timeout for the device from the MSEARCH response.
        """
        return self._cachecontrol

    @property
    def description(self) -> UpnpDevice1Device:
        """
            The UpnpDevice1Device device description.
        """
        desc = self._description
        return desc

    @property
    def embedded_device_descriptions(self) -> List[UpnpDevice1Device]:
        """
            Returns the list of descriptions for the embedded device that are declared for this root device.
        """
        dev_desc_list = []

        dev_desc = self._device_descriptions
        if dev_desc is not None:
            dev_desc_list = [devdesc for devdesc in dev_desc.values()]

        return dev_desc_list

    @property
    def embedded_devices(self) -> List[UpnpEmbeddedDevice]:
        """
            Returns the list of embedded devices that are declared for this root device.
        """
        devices_list = None

        devices = self._devices
        if devices is not None:
            devices_list = [dev for dev in devices]

        return devices_list

    @property
    def ext(self) -> str:
        """
            Returns the EXT header reported by the device in the MSEARCH response.
        """
        return self._ext

    @property
    def extra(self) -> dict:
        """
            Returns a dictionary that contains the extra thirdparty headers attached to UPNP MSEARCH responses.
        """
        return self._extra

    @property
    def IPAddress(self) -> str:
        """
            Returns the ip address of the root device that was found when the device responded with its MSEARCH response.
        """
        return self._ip_address

    @property
    def last_alive(self) -> datetime:
        """
            The datetime marker of the last time a UPNP alive or MSEARCH response was received
            from this device.
        """
        return self._last_alive

    @property
    def last_byebye(self) -> datetime:
        """
            The datetime marker of the last time a UPNP byebye was received from this device.
        """
        return self._last_byebye

    @property
    def MACAddress(self) -> Union[str, None]:
        """
            The MAC address of the device if specified in the device description or None.
        """
        macaddr = None
        desc = self.description
        if desc is not None:
            macaddr = desc.MACAddress
        return macaddr

    @property
    def mode(self) -> str:
        """
            The virtual mode assocatied with a device.  A device might present a different set of
            services depending on its current mode of operation.
        """
        return self._mode

    @property
    def modelName(self) -> Union[str, None]:
        """
            The model name of the device as found in the device description or None.
        """
        mname = None
        desc = self.description
        if desc is not None:
            mname = desc.modelName
        return mname

    @property
    def modelNumber(self )-> Union[str, None]:
        """
            The model number of the device as found in the device description or None.
        """
        mnumber = None
        desc = self.description
        if desc is not None:
            mnumber = desc.modelNumber
        return mnumber

    @property
    def primary_route(self):
        """
            Returns the primary rout for this UPNP devices.
        """
        return self._primary_route

    @property
    def routes(self) -> list:
        """
            A list of routes that can be used to communicate with the device.
        """
        return self._routes

    @property
    def server(self) -> str:
        """
            The 'SERVER' key specified in the MSEARCH response for this device.
        """
        return self._server

    @property
    def services(self) -> List[UpnpServiceProxy]:
        services_list = None

        services = self._services
        if services is not None:
            services_list = [svc for svc in services]

        return services_list

    @property
    def serialNumber(self) -> str:
        """
            The serial number number of the device as found in the device description.
        """
        desc = self.description
        return desc.serialNumber

    @property
    def specVersion(self) -> UpnpDevice1SpecVersion:
        """
            Returns the spec version of the device description.
        """
        return self._specVersion

    @property
    def USN(self) -> str:
        """
            The USN identifier specified in the MSEARCH response for this device.
        """
        return self._usn

    @property
    def USN_DEV(self) -> str:
        """
            The device component of the USN identifier specified in the MSEARCH response for this device.
        """
        return self._usn_dev

    @property
    def USN_CLS(self) -> str:
        """
            The class component of the USN identifier specified in the MSEARCH response for this device.
        """
        return self._usn_cls

    def get_custom_subscribe_headers(self) -> Optional[Dict[str, str]]:
        """
            An override-able method that allows for customization of the subscribe headers.
        """
        return None

    def initialize(self, coord_ref: weakref.ReferenceType, basedevice_ref: weakref.ReferenceType, extid: str, location: str, configinfo: dict, devinfo: dict):
        """
            Initializes the landscape device extension.

            :param coord_ref: A weak reference to the coordinator that is managing interactions through this
                              device extension.
            :param extid: A unique reference that can be used to identify this device via the coordinator even if its location changes.
            :param location: The location reference where this device can be found via the coordinator.
            :param configinfo: The UPNP configuration information associated with a device in the landscape configuration file.
            :param devinfo:  The device configuration information from the landscape configuration file.
        """
        # pylint: disable=arguments-differ

        LandscapeDeviceExtension.initialize(self, coord_ref, basedevice_ref, extid, location, configinfo)

        # We have to make a copy of the dictionary because we use
        # pop semantics to remove the items so we can pass on remaining
        # items to `_consume_upnp_extra`
        devinfo = devinfo.copy()

        self._cachecontrol = None
        if MSearchKeys.CACHE_CONTROL in devinfo:
            self._cachecontrol = devinfo.pop(MSearchKeys.CACHE_CONTROL)

        self._ext = None
        if MSearchKeys.EXT in devinfo:
            self._ext = devinfo.pop(MSearchKeys.EXT)

        self._server = None
        if MSearchKeys.SERVER in devinfo:
            self._server = devinfo.pop(MSearchKeys.SERVER)

        self._st = None
        if MSearchKeys.ST in devinfo:
            self._st = devinfo.pop(MSearchKeys.ST)

        self._usn = None
        if MSearchKeys.USN in devinfo:
            self._usn = devinfo.pop(MSearchKeys.USN)

        self._usn_dev = None
        if MSearchKeys.USN_DEV in devinfo:
            self._usn_dev = devinfo.pop(MSearchKeys.USN_DEV)

        self._usn_cls = None
        if MSearchKeys.USN_CLS in devinfo:
            self._usn_cls = devinfo.pop(MSearchKeys.USN_CLS)

        self._routes = [r for r in devinfo.pop(MSearchKeys.ROUTES).values()]

        self._primary_route = self._routes[0]

        self._consume_upnp_extra(devinfo)
        return

    def lookup_device(self, device_type: str) -> Union[UpnpEmbeddedDevice, None]:
        """
            Looks up an embedded device by its device type identifier.

            :param device_type: The device type identifier of the embedded device to find.

            :returns: The specified UpnpEmbeddedDevice or None
        """
        device = None

        self._device_lock.acquire()
        try:
            device = self._devices[device_type]
        finally:
            self._device_lock.release()

        return device

    def mark_alive(self):

        self._device_lock.acquire()
        try:
            self._last_alive = datetime.now()
            self._available = True
        finally:
            self._device_lock.release()

        return

    def mark_byebye(self):

        self.mark_unavailable(reason="byebye", byebye=True)

        return

    def mark_unavailable(self, reason: str="mark_unavailable", byebye: bool=False):
        """
            :param byebye: Indicates this method is being called from 'mark_byebye' as a result
                           of a byebye message being received.
        """

        self._device_lock.acquire()
        try:
            if byebye:
                self._last_byebye = datetime.now()

            if self._available:

                device_identity = self.basedevice.identity

                self._available = False

                subscription_type_list = [st for st in self._subscriptions.keys()]

                for svc_type in subscription_type_list:
                    sub_sid, _ = self._subscriptions[svc_type]

                    if sub_sid in self._sid_to_service_lookup:
                        svc_obj = self._sid_to_service_lookup[sub_sid]

                    # Call notify_byebye on each service object so we can mark all the
                    # references to variables as expired.
                    svc_obj.invalidate_subscription()

                    # Since we got a byebye, remove the subscription information
                    dbgmsg = "Removing subscription for device {} service {}, because of {}.".format(
                        device_identity, svc_type, reason
                    )
                    self._logger.debug(dbgmsg)
                    del self._subscriptions[svc_type]

        finally:
            self._device_lock.release()

        return

    def process_subscription_callback(self, sender_ip, sid: str, headers: dict, body: str):
        """
            This is used by the subscription callback thread to handoff work packets to the a worker thread
            and completes the processing of the subscription callback and routing to a UpnpRootDevice and
            UpnpServiceProxy instance.

            :param sender_ip: The IP address of the entity sending the update
            :param sid: The Subscription ID (sid) associated with the subscription callback
            :param headers: The HTTP headers contained in the callback response.
            :param body: The body content of the HTTP subscription callback.
        """
        # pylint: disable=unused-argument

        service = None

        self._device_lock.acquire()
        try:
            if sid in self._sid_to_service_lookup:
                service = self._sid_to_service_lookup[sid]
        finally:
            self._device_lock.release()

        if service is not None:
            docTree = ElementTree(xml_fromstring(body))

            psetNode = docTree.getroot()

            if psetNode is not None and psetNode.tag == "{%s}propertyset" % NS_UPNP_EVENT:
                propertyNodeList = psetNode.findall("{%s}property" % NS_UPNP_EVENT)

                service._update_event_variables(sender_ip, self._usn_dev, propertyNodeList) # pylint: disable=protected-access

        return

    def query_device_description(self):
        """
            Queries the UPnP device for its description.
        """
        desc = device_description_load(self._location)
        return desc


    def record_description(self, ip_addr: str, urlBase: str, manufacturer: str, modelName: str, docTree: ElementTree, devNode: Element, namespaces: str, upnp_recording: bool = False):
        """
            Called to record a description of a UPNP root device.

            :param urlBase: The base url as detailed in the description document or if not specified as determined by looking at the host information.
            :param manufacturer: The manufacturer of the device.
            :param modelName: The model name of the device.
            :param docTree: The XML document tree from the root of the device description.
            :param devNode: The 'device' element node from the device description.
            :param namespaces: A dictionary of namespaced to use when processing the XML document.
            :param upnp_recording: Force the recording of the device description and will overwrite existing device descriptions.
        """
        manufacturerNormalized = normalize_name_for_path(manufacturer)
        modelName = normalize_name_for_path(modelName)

        root_dev_dir = os.path.join(DIR_UPNP_SCAN_INTEGRATION_ROOTDEVICES, manufacturerNormalized)
        if not os.path.exists(root_dev_dir):
            os.makedirs(root_dev_dir)

        root_dev_def_file = os.path.join(root_dev_dir, modelName + ".xml")

        if upnp_recording:
            console_msg_lines = [
                "================================================================",
                " Recording Device: {}".format(ip_addr),
                "------------------------- Detail -------------------------------",
                "     Manufacturer: {}".format(manufacturer),
                "            Model: {}".format(modelName),
                "      Device File: {}".format(root_dev_def_file)
            ]
            console_msg = os.linesep.join(console_msg_lines)
            print(console_msg)

        if upnp_recording or not os.path.exists(root_dev_def_file):
            docNode = docTree.getroot()
            register_namespace('', namespaces[''])
            pretty_dev_content = xml_tostring(docNode, short_empty_elements=False)
            with open(root_dev_def_file, 'wb') as rddf:
                rddf.write(pretty_dev_content)

        indent = "    "
        next_indent = indent + "    "
        embDevList = devNode.find("deviceList", namespaces=namespaces)
        if embDevList is not None:
            if upnp_recording:
                print("{}[ Embedded Devices ]".format(indent))

            for embDevNode in embDevList:
                self._record_embedded_device(urlBase, manufacturerNormalized, embDevNode, namespaces, upnp_recording=upnp_recording, indent=next_indent)

            if upnp_recording:
                print("")

        svcList = devNode.find("serviceList", namespaces=namespaces)
        if svcList is not None:
            if upnp_recording:
                print("{}[ Services ]".format(indent))

            for svcNode in svcList:
                self._record_service( urlBase, manufacturerNormalized, svcNode, namespaces, upnp_recording=upnp_recording, indent=next_indent)

            if upnp_recording:
                print("")


        if upnp_recording:
            print()

        return

    def refresh_description(self, ipaddr: str, factory: "UpnpFactory", docNode: Element, namespaces=Optional[dict]):
        """
            Called by the UPNP coordinator to refresh the decription information for a device.

            :param ipaddr: IP address of the device to update the description for.
            :param factory: The UpnpFactory that can be used to instantiate devices.
            :param docNode: The root node of the description document.
            :param namespaces: A dictionary of namespaces to use when processing the device description.
        """
        try:
            if self._ip_address is not None and self._ip_address != ipaddr:
                self._logger.debug("UpnpRootDevice - IP Address change for device usn={} location={}.".format(
                    self._usn, self._ip_address
                ))

            self._ip_address = ipaddr

            specVerNode = docNode.find("specVersion", namespaces=namespaces)
            if specVerNode is not None:
                self._locked_process_version_node(specVerNode, namespaces=namespaces)

            baseURLNode = docNode.find("URLBase", namespaces=namespaces)
            if baseURLNode is not None:
                self._locked_process_urlbase_node(baseURLNode, namespaces=namespaces)

            url_parts = urlparse(self._location)
            self._host = url_parts.netloc

            # If urlBase was not set we need to try to use the schema and host as the urlBase
            if self._urlBase is None:
                self._urlBase = "%s://%s" % (url_parts.scheme, self._host)

            devNode = docNode.find("device", namespaces=namespaces)
            if devNode is not None:
                self._process_device_node(factory, devNode, namespaces=namespaces)

            self._enhance_device_detail()

        except Exception:
            err_msg = traceback.format_exc()
            print(err_msg)
            raise

        return

    def set_auto_subscribe(self, val, factory: Optional["UpnpFactory"] = None):
        self._auto_subscribe = val
        return

    def subscribe_to_all_services(self, factory: Optional["UpnpFactory"] = None):
        """
            The 'subscribe_to_all_services' method is overriden by derived objects in order
            to allow for bulk subscription to device services.

            ..note: Use with caution, if auto_subscribe is turned on, subscriptions are automatically
            created when the values of variables are attempted to be consumed via 'wait_for_value' and
            'wait_for_update'
        """

        for svc_name in self.SERVICE_NAMES:
            svc = self.lookup_service(self.MANUFACTURER, svc_name, allow_none=True, factory=factory)
            if svc is not None:
                try:
                    self.subscribe_to_events(svc)
                except:
                    errmsg = traceback.format_exc()
                    self._logger.debug(errmsg)
            else:
                errmsg = "Lookup service failed for mfg={} svc={}".format(self.MANUFACTURER, svc_name)
                self._logger.debug(errmsg)

        return

    def subscribe_to_events(self, service: UpnpServiceProxy, renew: bool=False, timeout: Optional[float] = None):
        """
            Creates a subscription to the event name specified and returns a
            UpnpEventVar object that can be used to read the current value for
            the given event.

            :param service: A :class:`akit.interop.upnp.services.upnpserviceproxy.UpnpServiceProxy`
                            for which to subscribe to events.
            :param timeout: The timeout to pass as a header when creating the subscription.
        """
        sub_sid = None
        sub_expires = None

        if len(service.SERVICE_EVENT_VARIABLES) > 0:

            service_type = service.SERVICE_TYPE

            new_subscription = False
            self._device_lock.acquire()
            try:
                if service_type not in self._subscriptions:
                    new_subscription = True
                    self._subscriptions[service_type] = (None, None)
                else:
                    sub_sid, sub_expires = self._subscriptions[service_type]

                    if sub_sid is None and sub_expires is None:
                        # The service was unsubscribed so this is the same as a new
                        # subscription.
                        new_subscription = True
            finally:
                self._device_lock.release()

            check_time = datetime.now() + timedelta(seconds=TIMEDELTA_RENEWAL_WINDOW)
            if sub_expires is not None and check_time > sub_expires:
                renew = True

            if new_subscription or renew:
                # If we created an uninitialized variable and added it to the subsciptions table
                # we need to statup the subsciption here.  If the startup process fails, we can
                # later remove the subscription from the subscription table.

                serviceManufacturer = normalize_name_for_path(self.MANUFACTURER)
                svckey = generate_extension_key(serviceManufacturer, service_type)
                service = self._services[svckey]

                subscribe_url = urljoin(self.URLBase, service.eventSubURL)
                subscribe_auth = ""

                coord = self._coord_ref()

                ifname = self._primary_route[MSearchRouteKeys.IFNAME]

                callback_url = coord.lookup_callback_url_for_interface(ifname)
                if callback_url is None:
                    errmsg = "No callback url found for ifname={}".format(ifname)
                    raise AKitRuntimeError(errmsg) from None

                headers = { "HOST": self._host, "User-Agent": AKitHttpHeaders.USER_AGENT, "CALLBACK": callback_url, "NT": "upnp:event"}
                if timeout is not None:
                    headers["TIMEOUT"] = "Seconds-%s" % timeout

                custom_headers = self.get_custom_subscribe_headers()
                if custom_headers is not None:
                    headers.update(custom_headers)

                resp = requests.request(
                    "SUBSCRIBE", subscribe_url, headers=headers, auth=subscribe_auth
                )

                if resp.status_code == 200:
                    #============================== Expected Response Headers ==============================
                    # SID: uuid:RINCON_7828CA09247C01400_sub0000000207
                    # TIMEOUT: Second-86400
                    # Server: Linux UPnP/1.0 Sonos/62.1-82260-monaco_dev (ZPS13)
                    # Connection: close
                    resp_headers = {k.upper(): v for k, v in resp.headers.items()}

                    nxtheader = None
                    try:
                        nxtheader = "SID"
                        sub_sid = resp_headers[nxtheader]

                        nxtheader = "TIMEOUT"
                        sub_timeout_str = resp_headers[nxtheader]
                    except KeyError as kerr:
                        errmsg = "Event subscription response was missing in %r header." % nxtheader
                        raise AKitCommunicationsProtocolError(errmsg) from kerr

                    sub_timeout = None
                    mobj = REGEX_SUBSCRIPTION_TIMEOUT.match(sub_timeout_str)
                    if mobj is not None:
                        timeout_str = mobj.groups()[0]
                        sub_timeout = 86400 if timeout_str == "infinite" else int(timeout_str)

                    if sub_sid is not None:
                        sub_expires = datetime.now() + timedelta(seconds=sub_timeout)
                        self._device_lock.acquire()
                        try:
                            self._sid_to_service_lookup[sub_sid] = service
                            self._subscriptions[service_type] = (sub_sid, sub_expires)
                        finally:
                            self._device_lock.release()

                        # Notify the coordinator which device has this subscription, we need to
                        # register the subscription ID so the coordinator can forward along
                        # notify message content
                        coord.register_subscription_for_device(sub_sid, self)
                    else:
                        self._device_lock.acquire()
                        try:
                            if service_type in self._subscriptions:
                                del self._subscriptions[service_type]
                        finally:
                            self._device_lock.release()

                else:
                    context_msg = "SUBSCRIBING at {}:".format(subscribe_url)
                    details = {
                        "SERVICE_KEY:": svckey,
                        "URL_BASE": self.URLBase,
                        "EVENT_SUB_URL": service.eventSubURL
                    }
                    raise_for_http_status(context_msg, resp, details=details)

        return sub_sid, sub_expires

    def unsubscribe_to_events(self, service: UpnpServiceProxy, subscription_id: Optional[str]=None):
        """
        """

        if len(service.SERVICE_EVENT_VARIABLES) > 0:
            if subscription_id is None:
                subscription_id = service.subscriptionId

            service._clear_subscription()

            subscribe_url = urljoin(self.URLBase, service.eventSubURL)
            subscribe_auth = ""
            headers = { "HOST": self._host, "User-Agent": AKitHttpHeaders.USER_AGENT, "SID": subscription_id}
                
            resp = requests.request(
                    "UNSUBSCRIBE", subscribe_url, headers=headers, auth=subscribe_auth
                )
            if resp.status_code != 200 and resp.status_code != 412:
                errmsg = "Error while unsubscribing from service."
                raise UpnpError(resp.status_code, errmsg)

        return

    def switchModes(self, mode: str):
        """
            Called to trigger the switching of a devices mode of operation.  This function is
            device specific and needs to be overloaded by custom UPNP device implementations.
        """
        self._mode = mode
        raise AKitNotOverloadedError("UpnpRootDevice.switchModes must be overloaded by custom UPNP devices.") from None

    def to_dict(self, brief=False) -> dict:
        """
            Returns a description of this device as a python dictionary.

            :param brief: A booliean value which indicates if a brief or full description is desired.

            :returns: The python dictionary description of the device.
        """
        dval = super(UpnpRootDevice, self).to_dict(brief=brief)
        dval["IPAddress"] = self.IPAddress
        dval["USN"] = self.USN
        dval["USN_DEV"] = self.USN_DEV
        dval["USN_CLS"] = self.USN_CLS
        return dval

    def to_json(self, brief=False) -> str:
        """
            Returns a description of this device in JSON format.

            :param brief: A booliean value which indicates if a brief or full description is desired.

            :returns: The JSON description of the device.
        """
        json_str = super(UpnpRootDevice, self).to_json(brief=brief)
        return json_str

    def _consume_upnp_extra(self, extrainfo: dict):
        """
            Called during the processing of MSEARCH responses for the device with
            a dictionary of any non-standard HTTP headers.

            :param extrainfo: A dictionary of extra headers found in the devices responses.  This
                              allows custom devices to overload this method and consume these extra
                              device response headers.
        """
        # pylint: disable=no-self-use

        self._extra = extrainfo
        return

    def _create_device_description_node(self, devNode: Element, namespaces: Optional[dict] = None):
        """
            Called in order for a device description to be created.  This could be overloaded by
            custom UPNP devices in order to create custome device description objects.

            :param devNode: The XML element of the root node of the device description document.
            :param namespaces: A dictionary of namespaces to use when processing the device description.
        """
        # pylint: disable=no-self-use
        dev_desc_node = UpnpDevice1Device(devNode, namespaces=namespaces)
        return dev_desc_node

    def _enhance_device_detail(self):
        """
            Can be implemented by custom devices in order to be able query for and enhance a devices
            information during the device update process.
        """
        # pylint: disable=no-self-use
        return

    def _locked_populate_embedded_device_descriptions(self, factory: "UpnpFactory", description: UpnpDevice1Device):
        """
            Called in order to process embedded device descriptions and to instantiate embedded device instances.

            NOTE: This method should be called with the device lock held.
        """
        for deviceInfo in description.deviceList:
            manufacturer = deviceInfo.manufacturer.strip()
            modelNumber = deviceInfo.modelNumber.strip()
            modelDescription = deviceInfo.modelDescription.strip()

            devkey = ":".join([manufacturer, modelNumber, modelDescription])

            if devkey not in self._device_descriptions:
                dev_inst = factory.create_embedded_device_instance(manufacturer, modelNumber, modelDescription)
                self._device_descriptions[devkey] = dev_inst

                self._locked_populate_services_descriptions(factory, deviceInfo)
            else:
                dev_inst = self._device_descriptions[devkey]

            dev_inst.update_description(self._host, self._urlBase, deviceInfo)
        return

    def _locked_process_urlbase_node(self, urlBaseNode: Element, namespaces: Optional[dict] = None):
        """
            Called in order to process the base url information from the device description.

            :param urlBaseNode: The XML Element node that contains the urlBaseNode information.
            :param namespaces: A dictionary of namespaces to use when processing the device description.

            NOTE: This method should be called with the device lock held.
        """
        # pylint: disable=unused-argument

        self._urlBase = urlBaseNode.text.rstrip("/")
        return

    def _locked_process_version_node(self, verNode: Element, namespaces: Optional[dict] = None):
        """
            Called in order to process the spec version information from the device description.

            :param verNode: The XML Element node of the version node.
            :param namespaces: A dictionary of namespaces to use when processing the device description.

            NOTE: This method should be called with the device lock held.
        """
        self._specVersion = UpnpDevice1SpecVersion(verNode, namespaces=namespaces)
        return

    def _matches_model_name(self, modelName: str) -> bool:
        """
            Method used to determine if the modelName passed matches the model name of this device.

            :param modelName: The modelName to evaluate for matching with this device.
        """
        matches = False
        if self.modelName == modelName:
            matches = True

        return matches

    def _matches_model_number(self, modelNumber: str)-> bool:
        """
            Method used to determine if the modelNumber passed matches the model number of this device.

            :param modelNumber: The modelNumber to evaluate for matching with this device.
        """
        matches = False
        if self.modelNumber == modelNumber:
            matches = True

        return matches

    def _process_device_node(self, factory: "UpnpFactory", devNode: Element, namespaces: Optional[dict] = None):
        """
            Method called for processing the 'device' node of the XML description of a device.

            :param factory: The UpnpFactory instance that is used to instantiate devices, embedded devices and service proxies.
            :param devNode: The XML Element node of the 'device' node.
            :param namespaces: A dictionary of namespaces to use when processing the device description.
        """
        description = self._create_device_description_node(devNode, namespaces=namespaces)

        self._device_lock.acquire()
        try:
            self._locked_populate_services_descriptions(factory, description)

            self._locked_populate_embedded_device_descriptions(factory, description)

            self._description = description
        finally:
            self._device_lock.release()

        return

    def _record_embedded_device(self, urlBase: str, manufacturer: str, embDevNode: Element, namespaces: Optional[dict] = None, upnp_recording: bool = False, indent="    "):
        """
            Method called for recording the description of an embedded device from a device description.

            :param manufacturer: The manufacturer of the device.
            :param embDevNode: The XML element node of the embedded device.
            :param namespaces: A dictionary of namespaces to use when processing the device description.
            :param upnp_recording: A boolean value indicating if the embedded device description should be recorded regardless of whether
                                    an existing description has been recorded.
        """
        # pylint: disable=no-self-use

        deviceTypeNode = embDevNode.find("deviceType", namespaces=namespaces)
        if deviceTypeNode is not None:
            deviceType = deviceTypeNode.text

            dyn_dev_dir = os.path.join(DIR_UPNP_SCAN_INTEGRATION_EMBEDDEDDEVICES, manufacturer)
            if not os.path.exists(dyn_dev_dir):
                os.makedirs(dyn_dev_dir)

            dyn_dev_filename = os.path.join(dyn_dev_dir, deviceType + ".xml")
            if upnp_recording or not os.path.exists(dyn_dev_filename):
                if upnp_recording:
                    dev_msg_lines = [
                        "{}Embedded Device Type: {}".format(indent, deviceType),
                        "{}Embedded Device File: {}".format(indent, dyn_dev_filename)
                    ]
                    dev_msg = os.linesep.join(dev_msg_lines)
                    print(dev_msg)

                pretty_sl_content = ""

                srcSvcListNode = embDevNode.find("serviceList", namespaces=namespaces)
                if srcSvcListNode is not None:
                    svcs_indent = indent + "    "
                    if upnp_recording:
                        print("{}[ Embedded Services ]".format(indent))

                    for svcNode in srcSvcListNode:
                        self._record_service( urlBase, manufacturer, svcNode, namespaces,
                                              upnp_recording=upnp_recording,
                                              indent=svcs_indent)

                    if upnp_recording:
                        print("")

                    register_namespace('', namespaces[''])
                    pretty_sl_content = xml_tostring(srcSvcListNode)

                with open(dyn_dev_filename, 'wb') as edf:
                    edf.write(b"<device>\n")
                    edf.write(pretty_sl_content)
                    edf.write(b"</device\n")

        return

    def _record_service(self, urlBase: str, manufacturer: str, svcNode: Element, namespaces: Optional[dict] = None, upnp_recording: bool = False, indent="    "):
        """
            Called to record the description of a device service.

            :param urlBase: The base url of the device which is used to request further information based on relative URLs from the device description.
            :param manufacturer: The manufacturer of the device.
            :param svcNode: The XML Element for the 'service' description node.
            :param namespaces: A dictionary of namespaces to use when processing the device description.
            :param upnp_recording: A boolean value indicating if the embedded device description should be recorded regardless of whether
                                    an existing description has been recorded.
        """
        # pylint: disable=broad-except

        serviceTypeNode = svcNode.find("serviceType", namespaces=namespaces)
        scpdUrlNode = svcNode.find("SCPDURL", namespaces=namespaces)
        if serviceTypeNode is not None and scpdUrlNode is not None:

            serviceType = serviceTypeNode.text

            scpdUrl = scpdUrlNode.text
            if urlBase is not None:
                scpdUrl = urlBase.rstrip("/") + "/" + scpdUrl.lstrip("/")

            dyn_service_dir = os.path.join(DIR_UPNP_SCAN_INTEGRATION_SERVICES, manufacturer)
            if not os.path.exists(dyn_service_dir):
                os.makedirs(dyn_service_dir)

            dyn_svc_filename = os.path.join(dyn_service_dir, "%s.xml" % (serviceType,))
            if upnp_recording or not os.path.exists(dyn_svc_filename):
                if upnp_recording:
                    svc_msg_lines = [
                        "{}Service Type: {}".format(indent, serviceType),
                        "{} Service URL: {}".format(indent, scpdUrl),
                        "{}Service File: {}".format(indent, dyn_svc_filename)
                    ]
                    svc_msg = os.linesep.join(svc_msg_lines)
                    print(svc_msg)

                try:
                    resp = requests.get(scpdUrl)
                    if resp.status_code == 200:
                        svc_content = resp.content
                        with open(dyn_svc_filename, 'wb') as sdf:
                            sdf.write(svc_content)
                    else:
                        self._logger.warn("Unable to retrieve service description for manf=%s st=%s url=%s" % (manufacturer, serviceType, scpdUrl))
                except Exception:
                    self._logger.exception("Exception while retreiving service description.")

        return

    def __str__(self):
        """
            Returns a brief description of the device as a string.
        """
        rtnstr = "%s: USN:%s MAC=%s IP=%s" % (self.modelName, self.USN, self.MACAddress, self.IPAddress)
        return rtnstr



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import inspect
import threading
import time

from akit.environment.variables import AKIT_VARIABLES
from akit.xlogging.foundations import getAutomatonKitLogger
from akit.wellknownports import PORT_DEBUGPY_ASSISTANT

logger = getAutomatonKitLogger()

class DEBUGGER:
    PDB = 'pdb'
    DEBUGPY = 'debugpy'

DEFAULT_DEBUG_PORT = 5678

class WELLKNOWN_BREAKPOINTS:
    TEST_DISCOVERY = "test-discovery"
    TESTRUN_START = "testrun-start"

def in_vscode_debugger():
    invscode = False
    for frame in inspect.stack():
        if frame[1].endswith("pydevd.py"):
            invscode = True
            break
    return invscode

def debugger_wellknown_breakpoint_entry(breakpoint_name: str):

    debugger = AKIT_VARIABLES.AKIT_DEBUGGER
    breakpoints = AKIT_VARIABLES.AKIT_BREAKPOINTS

    if breakpoints is not None:
        if breakpoint_name in breakpoints:
            if debugger == DEBUGGER.PDB:
                # The debug flag was passed on the commandline so we break here.'.format(current_indent))
                import pdb
                pdb.set_trace()
            elif debugger == DEBUGGER.DEBUGPY:
                logger.info("Waiting for debugger on port={}".format(DEFAULT_DEBUG_PORT))

                invscode = in_vscode_debugger()

                # The remote debug flag was passed on the commandline so we break here.'.format(current_indent))
                import debugpy
                if invscode:
                    debugpy.listen(("0.0.0.0", DEFAULT_DEBUG_PORT))
                    debugpy.wait_for_client()
                debugpy.breakpoint()
    return

def debugger_wellknown_breakpoint_code_append(breakpoint_name: str, code_lines: list, current_indent: str):

    debugger = AKIT_VARIABLES.AKIT_DEBUGGER
    breakpoints = AKIT_VARIABLES.AKIT_BREAKPOINTS

    if breakpoints is not None:
        if breakpoint_name in breakpoints:
            if debugger == DEBUGGER.PDB:
                code_lines.append('')
                code_lines.append('{}# The debug flag was passed on the commandline so we break here.'.format(current_indent))
                code_lines.append('{}import pdb'.format(current_indent))
                code_lines.append('{}pdb.set_trace()'.format(current_indent))
            elif debugger == DEBUGGER.DEBUGPY:
                invscode = in_vscode_debugger()

                code_lines.append('')
                code_lines.append('{}# The remote debug flag was passed on the commandline so we break here.'.format(current_indent))
                code_lines.append('{}import debugpy'.format(current_indent))
                if invscode:
                    code_lines.append('{}debugpy.listen(("0.0.0.0", {}))'.format(current_indent, DEFAULT_DEBUG_PORT))
                    code_lines.append('{}debugpy.wait_for_client()'.format(current_indent))
                code_lines.append('{}debugpy.breakpoint()'.format(current_indent))

                logger.info("Waiting for debugger on port={}".format(DEFAULT_DEBUG_PORT))

    return


class DebugPyAssistant():
    """
        The DebugPyAssistant is used to setup a daemon thread that creates a remote debug
        endpoint that can optionally connected to in order to remote debug a process running
        in the automation environment.
    """

    def __init__(self, name="DebugPyAssistant", *, endpoint=("0.0.0.0", PORT_DEBUGPY_ASSISTANT)):
        self._endpoint = endpoint
        self._running = True

        self._debug_thread = threading.Thread(name=name, target=self._entry_debug_server, daemon=True)
        self._debug_thread.start()
        return

    @property
    def endpoint(self):
        return self._endpoint

    @property
    def thread(self):
        return self._debug_thread

    def _entry_debug_server(self):

        import debugpy
        debugpy.listen(self._endpoint)
        
        while self._running:
            debugpy.wait_for_client()
            debugpy.breakpoint()

            # While we are connected to the debugger, have the debug assistant
            # thread loop
            while debugpy.is_client_connected():
                time.sleep(2)

        return




__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Tuple

import multiprocessing
import multiprocessing.managers

from akit.networking.simplewebserver import SimpleWebServer

class SimpleWebServerManager(multiprocessing.managers.BaseManager):
    """
        This is a process manager used for creating a :class:`SimpleWebServer`
        in a remote process that can be communicated with via a proxy.
    """

SimpleWebServerManager.register("SimpleWebServer", SimpleWebServer)

def spawn_webserver_process(address: Tuple[str, int], rootdir: str, protocol: str="HTTP/1.0") -> Tuple[SimpleWebServerManager, SimpleWebServer]:
    srvr_mgr = SimpleWebServerManager()
    srvr_mgr.start()
    wsvr_proxy = srvr_mgr.SimpleWebServer(address, rootdir, protocol)
    return srvr_mgr, wsvr_proxy



"""
.. module:: dnsaddress
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Contains the DnsAddress record type.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import TYPE_CHECKING

import socket

from akit.networking.interfaces import is_ipv6_address

from akit.interop.dns.dnsrecord import DnsRecord

if TYPE_CHECKING:
    from akit.interop.dns.dnspacketwriter import DnsPacketWriter

class DnsAddress(DnsRecord):
    """
        A DNS address record
    """

    def __init__(self, name: str, rtype: int, rclass: int, ttl: int, address: bytes) -> None:
        DnsRecord.__init__(self, name, rtype, rclass, ttl)
        self._address = address
        return

    @property
    def address(self):
        return self._address

    def write(self, out: 'DnsPacketWriter') -> None:
        """
            Used in constructing an outgoing packet
        """
        out.write_string(self._address)
        return

    def __eq__(self, other: DnsRecord) -> bool:
        """
            Tests equality on priority, weight, port and server
        """
        # Call DnsRecord equality operator first because it can eliminate alot of equality checks early
        iseq = False
        if DnsRecord.__eq__(self, other):
            iseq = self._address == other._address
        return iseq

    def __ne__(self, other: DnsRecord) -> bool:
        """
            Non-equality test
        """
        isneq = not self.__eq__(other)
        return isneq

    def __str__(self) -> str:
        """
            String representation
        """
        strval = None

        try:
            strval = self.as_dns_string(
                socket.inet_ntop(
                    socket.AF_INET6 if is_ipv6_address(self._address) else socket.AF_INET, self._address
                )
            )
        except Exception:  # TODO stop catching all Exceptions
            strval = self.as_dns_string(str(self._address))

        return strval


__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import enum
import socket
import struct
import telnetlib

COMMAND_NAMES = {
    telnetlib.BRK: "BRK",
    telnetlib.DM: "DM",
    telnetlib.DO: "DO",
    telnetlib.DONT: "DONT",
    telnetlib.GA: "GA",
    telnetlib.IAC: "IAC",
    telnetlib.NOP: "NOP",
    telnetlib.SB: "SB",
    telnetlib.SE: "SE",
    telnetlib.WILL: "WILL",
    telnetlib.WONT: "WONT"
}

OPTION_NAMES = {
    telnetlib.BINARY: "BINARY",
    telnetlib.ECHO: "ECHO",
    telnetlib.RCP: "RCP",
    telnetlib.SGA: "SGA",
    telnetlib.NAMS: "NAMS",
    telnetlib.STATUS: "STATUS",
    telnetlib.TM: "TM",
    telnetlib.RCTE: "RCTE",
    telnetlib.NAOL: "NAOL",
    telnetlib.NAOP: "NAOP",
    telnetlib.NAOCRD: "NAOCRD",
    telnetlib.NAOHTS: "NAOHTS",
    telnetlib.NAOHTD: "NAOHTD",
    telnetlib.NAOFFD: "NAOFFD",
    telnetlib.NAOVTS: "NAOVTS",
    telnetlib.NAOVTD: "NAOVTD",
    telnetlib.NAOLFD: "NAOLFD",
    telnetlib.XASCII: "XASCII",
    telnetlib.LOGOUT: "LOGOUT",
    telnetlib.BM: "BM",
    telnetlib.DET: "DET",
    telnetlib.SUPDUP: "SUPDUP",
    telnetlib.SUPDUPOUTPUT: "SUPDUPOUTPUT",
    telnetlib.SNDLOC: "SNDLOC",
    telnetlib.TTYPE: "TTYPE",
    telnetlib.EOR: "EOR",
    telnetlib.TUID: "TUID",
    telnetlib.OUTMRK: "OUTMRK",
    telnetlib.TTYLOC: "TTYLOC",
    telnetlib.VT3270REGIME: "VT3270REGIME",
    telnetlib.X3PAD: "X3PAD",
    telnetlib.NAWS: "NAWS",
    telnetlib.TSPEED: "TSPEED",
    telnetlib.LFLOW: "LFLOW",
    telnetlib.LINEMODE: "LINEMODE",
    telnetlib.XDISPLOC: "XDISPLOC",
    telnetlib.OLD_ENVIRON: "OLD_ENVIRON",
    telnetlib.AUTHENTICATION: "AUTHENTICATION",
    telnetlib.ENCRYPT: "ENCRYPT",
    telnetlib.NEW_ENVIRON: "NEW_ENVIRON",
    telnetlib.TN3270E: "TN3270E",
    telnetlib.XAUTH: "XAUTH",
    telnetlib.CHARSET: "CHARSET",
    telnetlib.RSP: "RSP",
    telnetlib.COM_PORT_OPTION: "COM_PORT_OPTION",
    telnetlib.SUPPRESS_LOCAL_ECHO: "SUPPRESS_LOCAL_ECHO",
    telnetlib.TLS: "TLS",
    telnetlib.KERMIT: "KERMIT",
    telnetlib.SEND_URL: "SEND_URL",
    telnetlib.FORWARD_X: "FORWARD_X",
    telnetlib.PRAGMA_LOGON: "PRAGMA_LOGON",
    telnetlib.SSPI_LOGON: "SSPI_LOGON",
    telnetlib.PRAGMA_HEARTBEAT: "PRAGMA_HEARTBEAT"
}

class TelnetOptionState(enum.IntEnum):
    UNSET = 0
    REQUESTED = 1
    ACTIVE = 2
    INACTIVE = 3
    REALLY_INACTIVE = 4


class TelnetOption:
    """
        Manages the state of a telnet option for a telnet session.
    """
    def __init__(self, name, option, send_yes, send_no, ack_yes, ack_no, initial_state, activation_callback=None):
        self._name = name
        self._option = option
        self._send_yes = send_yes
        self._send_no = send_no
        self._ack_yes = ack_yes
        self._ack_no = ack_no
        self._state = initial_state
        self._active = False
        self._activation_callback = activation_callback
        return

    def negotiate_option(self, command):
        """
            A DO/DONT/WILL/WONT was received for this option, update state and answer when needed.
        """
        if command == self._ack_yes:
            pass
        elif command == self._ack_no:
            pass
        return


class SerialTelnet(telnetlib.Telnet):

    def __init__(self, host=None, port=0, baudrate=None, echo=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT):

        super(SerialTelnet, self).__init__(host=host, port=port, timeout=timeout)

        self._baudrate = 115200
        if baudrate is not None:
            self._baudrate = baudrate

        self._echo = False
        if echo is not None:
            self._echo = echo

        self.set_option_negotiation_callback(self._callback_negotiate_options)
        return

    def open(self, host, port=0, baudrate=None, echo=None, timeout=socket._GLOBAL_DEFAULT_TIMEOUT):
        """Connect to a host.

        The optional second argument is the port number, which
        defaults to the standard telnet port (23).

        Don't try to reopen an already connected instance.
        """

        if baudrate is not None:
            self._baudrate = baudrate

        if echo is not None:
            self._echo = echo

        super(SerialTelnet, self).open(host=host, port=port, timeout=timeout)

        return

    def negotiate_options(self):
        self._send_option(telnetlib.DONT, telnetlib.ECHO)
        return

    def _send_option(self, action, option):
        pkstr = struct.pack(">ccc", telnetlib.IAC, action, option)
        tsocket = self.get_socket()
        self.sock.sendall(pkstr)
        return

    def _callback_negotiate_options(self, tsocket, command, options):

        cmdstr = COMMAND_NAMES[command]
        optstr = OPTION_NAMES[options]

        print("cmd=%s opt=%s" % (cmdstr, optstr))

        if command == telnetlib.DO:
            if options == telnetlib.TTYPE:
                # Promise we will send a terminal type
                self._send_option(telnetlib.WONT, telnetlib.TTYPE)
            elif options == telnetlib.ECHO:
                # Promise we will echo
                self._send_option(telnetlib.DONT, telnetlib.ECHO)
            elif options == telnetlib.NAWS:
                # Promise we will send our window size
                self._send_option(telnetlib.WONT, telnetlib.NAWS)
            elif options == telnetlib.TSPEED:
                # Promise we will send a terminal speed
                self._send_option(telnetlib.WILL, telnetlib.NAWS)
            elif options == telnetlib.XDISPLOC:
                # Tell the server we will not send the x-display terminal
                self._send_option(telnetlib.WONT, telnetlib.XDISPLOC)
        elif command == telnetlib.SB:
            if options == telnetlib.NEW_ENVIRON:
                print("Asked for environment.")
        # cmd=WILL opt=03
        # cmd=WILL opt=01
        # cmd=DONT opt=01
        # cmd=DO   opt=00
        # cmd=WONT opt=01

        return


if __name__ == "__main__":
    stn = SerialTelnet()
    stn.open(host="192.168.1.17", port=3333)
    stn.read_very_eager()

    stn.negotiate_options()

    stn.read_very_eager()

    stn.interact()

    print("blah")




"""
.. module:: xtime
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains framework time related functions which extend the functionality to
               the python :module:`time` and  :module:`datetime` modules.

.. note:: The modules that are named `xsomething` like this module are prefixed with an `x` character to
          indicate they extend the functionality of a base python module and the `x` is pre-pended to
          prevent module name collisions with python modules.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import time

from datetime import datetime

FORMAT_DATETIME = "%Y-%m-%dT%H%M!%S"

def current_time_millis() -> float:
    """
        Current system time in milliseconds

        :returns: Time in milliseconds
    """
    now_ms = time.time() * 1000
    return now_ms

def format_time_with_fractional(tsecs: float) -> str:
    """
        Format the time in seconds as a fractional in seconds.

        :param tsecs: Time in seconds as a float.

        :returns: Formatted time in (seconds).(fractions of seconds)
    """
    sec_comp = int(tsecs)
    frac_comp = (tsecs - sec_comp) * 1000
    dtstr = "%s.%06d" % (time.strftime(FORMAT_DATETIME, time.gmtime(sec_comp)), frac_comp)
    return dtstr

def parse_datetime(dtstr: str, datetime_format: str=FORMAT_DATETIME) -> datetime:
    """
        Parses a date time from string and includes the microseconds component.

        :param dtstr: The date in the form of a date time string.
        :param datetime_format: The format string to when parsing the datetime string.

        :returns: The datetime from the parsed string.
    """
    microsecs = 0

    if dtstr.find(".") > 0:
        dtstr, msecstr = dtstr.split(".")
        microsecs = int(msecstr)

    dtobj = datetime(*time.strptime(dtstr, datetime_format)[:6], microsecs)

    return dtobj



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import NamedTuple

class StationInfo(NamedTuple):
    mac: str
    inactive_time: int
    rx_bytes: int
    rx_packets: int
    rx_drop_misc: int
    rx_bitrate: int
    tx_bytes: int
    tx_packets: int
    tx_retries: int
    tx_failed: int
    tx_bitrate: int
    signal: int
    signal_avg: int
    exp_throughput: int
    count_authorized: int
    count_authenticated: int
    count_associated: int
    wmm: int
    mfp: int
    tdls_peer: str
    dtim_period: str
    beacon_interval: float
    preamble: int
    short_preamble: int
    short_slot_time: int
    connected_time: int



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class QPlay1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'QPlay1' service.
    """

    SERVICE_MANUFACTURER = 'SonosInc'
    SERVICE_TYPE = 'urn:schemas-tencent-com:service:QPlay:1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {}

    def action_QPlayAuth(self, Seed, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the QPlayAuth action.

            :returns: "Code", "MID", "DID"
        """
        arguments = {
            "Seed": Seed,
        }

        out_params = self.call_action("QPlayAuth", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("Code", "MID", "DID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args



"""

    NOTE: This is a code generated file.  This file should not be edited directly.
"""



from akit.aspects import AspectsUPnP, DEFAULT_UPNP_ASPECTS

from akit.extensible import LoadableExtension
from akit.interop.upnp.services.upnpserviceproxy import UpnpServiceProxy

class WANIPv6FirewallControl1ServiceProxy(UpnpServiceProxy, LoadableExtension):
    """
        This is a code generated proxy class to the 'urn:schemas-upnp-org:service:WANIPv6FirewallControl:1' service.
    """

    SERVICE_MANUFACTURER = 'UPnP'
    SERVICE_TYPE = 'WANIPv6FirewallControl1'

    SERVICE_DEFAULT_VARIABLES = {}

    SERVICE_EVENT_VARIABLES = {
        "FirewallEnabled": { "data_type": "boolean", "default": None, "allowed_list": None},
        "InboundPinholeAllowed": { "data_type": "boolean", "default": None, "allowed_list": None},
    }

    def action_AddPinhole(self, RemoteHost, RemotePort, InternalClient, InternalPort, Protocol, LeaseTime, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the AddPinhole action.

            :returns: "UniqueID"
        """
        arguments = {
            "RemoteHost": RemoteHost,
            "RemotePort": RemotePort,
            "InternalClient": InternalClient,
            "InternalPort": InternalPort,
            "Protocol": Protocol,
            "LeaseTime": LeaseTime,
        }

        out_params = self.call_action("AddPinhole", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("UniqueID",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_CheckPinholeWorking(self, UniqueID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the CheckPinholeWorking action.

            :returns: "IsWorking"
        """
        arguments = {
            "UniqueID": UniqueID,
        }

        out_params = self.call_action("CheckPinholeWorking", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("IsWorking",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_DeletePinhole(self, UniqueID, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the DeletePinhole action.
        """
        arguments = {
            "UniqueID": UniqueID,
        }

        self.call_action("DeletePinhole", arguments=arguments, aspects=aspects)

        return

    def action_GetFirewallStatus(self, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetFirewallStatus action.

            :returns: "FirewallEnabled", "InboundPinholeAllowed"
        """
        arguments = { }

        out_params = self.call_action("GetFirewallStatus", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("FirewallEnabled", "InboundPinholeAllowed",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetOutboundPinholeTimeout(self, RemoteHost, RemotePort, InternalClient, InternalPort, Protocol, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetOutboundPinholeTimeout action.

            :returns: "OutboundPinholeTimeout"
        """
        arguments = {
            "RemoteHost": RemoteHost,
            "RemotePort": RemotePort,
            "InternalClient": InternalClient,
            "InternalPort": InternalPort,
            "Protocol": Protocol,
        }

        out_params = self.call_action("GetOutboundPinholeTimeout", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("OutboundPinholeTimeout",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_GetPinholePackets(self, UniqueID, *, extract_returns=True, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the GetPinholePackets action.

            :returns: "PinholePackets"
        """
        arguments = {
            "UniqueID": UniqueID,
        }

        out_params = self.call_action("GetPinholePackets", arguments=arguments, aspects=aspects)

        rtn_args = out_params
        if extract_returns:
            rtn_args = [out_params[k] for k in ("PinholePackets",)]
            if len(rtn_args) == 1:
                rtn_args = rtn_args[0]

        return rtn_args

    def action_UpdatePinhole(self, UniqueID, NewLeaseTime, *, aspects:AspectsUPnP=DEFAULT_UPNP_ASPECTS):
        """
            Calls the UpdatePinhole action.
        """
        arguments = {
            "UniqueID": UniqueID,
            "NewLeaseTime": NewLeaseTime,
        }

        self.call_action("UpdatePinhole", arguments=arguments, aspects=aspects)

        return



__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

from typing import Tuple, Union

import enum
import errno
import socket
import struct
import threading

from akit.exceptions import AKitRuntimeError
from akit.networking.constants import MDNS_GROUP_ADDR, MDNS_PORT
from akit.networking.multicast import create_multicast_socket_for_iface

class DnsAgent:

    def __init__(self, ifname):
        self._ifname = ifname
        self._thread = None
        self._running = False
        return

    def start(self):
        sgate = threading.Event()
        self._thread = threading.Thread(name="DnsAgent - Monitor", target=self._thread_entry_monitor, daemon=True, args=(sgate,))
        return

    def _thread_entry_monitor(self, sgate):

        sock = None
        try:
            sock = create_multicast_socket_for_iface(MDNS_GROUP_ADDR, self._ifname, MDNS_PORT, socket.AF_INET, ttl=1)

            while self._running:
                request, addr = sock.recvfrom(1024)

                response, addr = sock.recvfrom(1024)
                print("PACKET: %s" % addr[0])
                response_lines = response.decode("utf-8").splitlines()
                for rline in response_lines:
                    print("    %s" % rline)
                print()

        finally:
            if sock is not None:
                sock.close()

        return


"""
.. module:: clients
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: The clients package contains couplings that are utilized to connect
               tests with the requirements and functionality to interoperate with
               linux and windows clients.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"



"""
.. module:: scopemonitor
    :platform: Darwin, Linux, Unix, Windows
    :synopsis: Module which contains the :class:`ScopeMonitor` object which monitors thread entrapment
               withing a specified scope.

.. note:: The modules that are named `xsomething` like this module are prefixed with an `x` character to
          indicate they extend the functionality of a base python module and the `x` is pre-pended to
          prevent module name collisions with python modules.

.. moduleauthor:: Myron Walker <myron.walker@gmail.com>
"""

__author__ = "Myron Walker"
__copyright__ = "Copyright 2020, Myron W Walker"
__credits__ = []
__version__ = "1.0.0"
__maintainer__ = "Myron Walker"
__email__ = "myron.walker@gmail.com"
__status__ = "Development" # Prototype, Development or Production
__license__ = "MIT"

import os
import threading
import time
import uuid

from datetime import datetime, timedelta

from akit.timeouts import TimeoutContext

from akit.xcollections import insert_into_ordered_list_ascending
from akit.xformatting import split_and_indent_lines
from akit.xlogging.foundations import getAutomatonKitLogger

DEFAULT_MONITORED_SCOPE_NOTIFY_DELAY = timedelta(seconds=60)

logger = getAutomatonKitLogger()

class MonitoredScope:
    """
        The :class:`MonitoredScope` object is utilized in order to provide monitoring on threads
        that are entering sections of code that might have a tendency to block and cause disruptions
        in automation processes.  The monitored scope allows for the creation of scopes of code execution
        that can capture contextual information and hand it to a monitor thread that can then log an error
        message if the thread entering the monitored context does not return from the context with a
        specified period of time.

        This solves the problem of reducing log clutter by not having to log at the entry of problematic
        sections of code, but delay the logging until the thread has failed to exit in a timely manner and
        ensure the logging can happen by passing the work off to another thread that is running in a safer
        context.
    """

    ERROR_COMPARISON_TYPE_MESSAGE = "Comparison is only support between two 'ScopeMonitor' objects."

    def __init__(self, label, message, timeout_ctx: TimeoutContext, notify_delay: timedelta=DEFAULT_MONITORED_SCOPE_NOTIFY_DELAY):
        self._id = str(uuid.uuid4())
        self._label = label
        self._message = message
        self._timeout_ctx = timeout_ctx
        self._notify_delay = notify_delay

        self._diag_func = None
        self._diag_args = None
        self._diag_kwargs = None

        self._exited = False
        self._triggered = False
        return

    def __enter__(self):
        """
        """
        global global_scope_monitor

        if global_scope_monitor is None:
            global_scope_monitor = ScopeMonitor()

        self._timeout_ctx.mark_begin()

        global_scope_monitor.register_monitor(self)

        return self

    def __exit__(self, ex_type, ex_inst, ex_tb):
        """
        """
        self._exited = True

        return False

    def __eq__(self, other: "ScopeMonitor"):
        """
            Perform comparison between ScopeMonitor(left) == ScopeMonitor(right)
        """
        if not isinstance(other, ScopeMonitor):
            raise ValueError(self.ERROR_COMPARISON_TYPE_MESSAGE)

        return self._timeout_ctx.end_time == other._timeout_ctx.end_time

    def __ge__(self, other: "ScopeMonitor"):
        """
            Perform comparison between ScopeMonitor(left) >= ScopeMonitor(right)
        """
        if not isinstance(other, ScopeMonitor):
            raise ValueError(self.ERROR_COMPARISON_TYPE_MESSAGE)

        return self._timeout_ctx.end_time >= other._timeout_ctx.end_time

    def __gt__(self, other: "ScopeMonitor"):
        """
            Perform comparison between ScopeMonitor(left) > ScopeMonitor(right)
        """
        if not isinstance(other, ScopeMonitor):
            raise ValueError(self.ERROR_COMPARISON_TYPE_MESSAGE)

        return self._timeout_ctx.end_time > other._timeout_ctx.end_time

    def __le__(self, other: "ScopeMonitor"):
        """
            Perform comparison between ScopeMonitor(left) <= ScopeMonitor(right)
        """
        if not isinstance(other, ScopeMonitor):
            raise ValueError(self.ERROR_COMPARISON_TYPE_MESSAGE)

        return self._timeout_ctx.end_time <= other._timeout_ctx.end_time

    def __lt__(self, other: "ScopeMonitor"):
        """
            Perform comparison between ScopeMonitor(left) < ScopeMonitor(right)
        """
        if not isinstance(other, ScopeMonitor):
            raise ValueError(self.ERROR_COMPARISON_TYPE_MESSAGE)

        return self._timeout_ctx.end_time < other._timeout_ctx.end_time

    def __ne__(self, other: "ScopeMonitor"):
        """
            Perform comparison between ScopeMonitor(left) != ScopeMonitor(right)
        """
        if not isinstance(other, ScopeMonitor):
            raise ValueError(self.ERROR_COMPARISON_TYPE_MESSAGE)

        return self._timeout_ctx.end_time != other._timeout_ctx.end_time

    @property
    def exited(self):
        """
            Return true if the thread that entered the MonitoredScope has exited.
        """
        return self._exited

    @property
    def expired(self):
        """
            Returns true if the MonitoredScope instance has not been exited by the calling
            thread before it has expired.
        """
        is_expired = False
        now = datetime.now()
        if now > (self._timeout_ctx.end_time + self._notify_delay):
            is_expired = True
        return is_expired

    @property
    def id(self):
        """
            The unique identifier for this MonitoredScope instance.
        """
        return self._id

    @property
    def label(self):
        """
            The human readable label has been assigned to identify this scope.
        """
        return self._label

    @property
    def message(self):
        """
            The message that will be logged if this MonitoredScope is not exited before
            the timemout has expired.
        """
        return self._message

    def set_diagnostic(self, diagnostic_function, *args, **kwargs):
        """
            Sets the diagnostic function and its associated args which will be run if
            the :class:`MonitoredScope` is not exited before it has expired.
        """
        self._diag_func = diagnostic_function
        self._diag_args = args
        self._diag_kwargs = kwargs
        return

    def trigger_notification(self):
        """
            On the first trigger of a notification.  Runs the diagnostic function, logs the label and
            message along with the information collected by the diagnostic function.
        """
        if not self._triggered:
            self._triggered = True
            if self.expired:

                errlines = [
                    "MonitoredScope({}): Timeout waiting for thread to exit monitored scope.".format(self._label),
                    "MESSAGE: {}".format(self._message),

                ]

                if  self._diag_func:
                    errlines.append("DIAGNOSTIC:")

                    diagmsg = self._diag_func(*self._diag_args, **self._diag_kwargs)
                    errlines.extend(split_and_indent_lines(diagmsg, 1))

                errmsg = os.linesep.join(errlines)
                logger.error(errmsg)

        return

class ScopeMonitor:
    """
        The :class:`ScopeMonitor` object is utilized to provide monitoring of threads that are entering
        sections of code that might have a tendency to block and cause disruptions in automation processes.

        The :class:`ScopeMonitor` takes pre-dictive work packets for logging from threads entering critical
        sections of code monitored by :class:`MonitoredScope` instances used in a 'with' statement.

        ..code-block:: python

            with MonitoredScope("RunCommand", "Running command on cluster node (%s)" % nodeip) as mscope:
                cluster.run_cmd(1, "echo blah")

        The use of a :class:`MonitoredScope` remove the necessity to log prior to entering a critical
        section of code but allows log entrys to be pre-emptively handed off to the :class:`ScopeMonitor`
        thread for contingent processing should a the thread fail to return from the critical section of
        code in a timely manner.
    """

    SCOPE_MONITOR_INTERVAL = 5

    instance = None
    initialized = False

    def __new__(cls, *_args, **_kwargs):
        """
            Constructs new instances of the :class:`ScopeMonitor` object. The
            :class:`ScopeMonitor` object is a singleton so following instantiations
            of the object will reference the existing singleton
        """

        if cls.instance is None:
            cls.instance = super(ScopeMonitor, cls).__new__(cls)
        return cls.instance

    def __init__(self):
        thisType = type(self)
        if not thisType.initialized:
            thisType.initialized = True

            self._monitors = []
            self._monitors_lock = threading.RLock()

            self._timer = threading.Timer(self.SCOPE_MONITOR_INTERVAL, self._monitor_tick)
        return

    def register_monitor(self, monitor: MonitoredScope):
        """
            Register a monitor context with the :class:`ScopeMonitor` singleton.

            :param monitor: The monitor context to add to the the list of monitored scopes.
        """

        self._monitors_lock.acquire()
        try:
            insert_into_ordered_list_ascending(self._monitors, monitor)
        finally:
            self._monitors_lock.release()

        return

    def _monitor_tick(self):

        self._monitors_lock.acquire()
        try:
            if len(self._monitors) > 0:
                while True:
                    firstMonitor = self._monitors[0]
                    if firstMonitor.expired:
                        firstMonitor.trigger_notification()
                        del self._monitors[0]
                    else:
                        break
        finally:
            self._monitors_lock.release()

        return


global_scope_monitor = None
